#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import json
import os
import uuid
from typing import Dict, Tuple, Union, cast

from apispec import APISpec  # type: ignore

from apispec_webframeworks.flask import FlaskPlugin  # type: ignore

from dataclasses_jsonschema.apispec import DataclassesPlugin

from flask import Flask, Response, jsonify, request

from flask_cors import CORS  # type: ignore

from flask_swagger_ui import get_swaggerui_blueprint  # type: ignore

import arcor2
from arcor2.data import common, object_type
from arcor2.helpers import camel_case_to_snake_case
from arcor2.rest import convert_keys

PORT = int(os.getenv("ARCOR2_SCENE_SERVICE_MOCK_PORT", 5013))
SERVICE_NAME = "ARCOR2 Scene Service Mock"


# Create an APISpec
spec = APISpec(
    title=SERVICE_NAME,
    version=arcor2.version(),
    openapi_version="3.0.2",
    plugins=[FlaskPlugin(), DataclassesPlugin()],
)

app = Flask(__name__)
CORS(app)


def get_id() -> int:
    return uuid.uuid4().int


RespT = Union[Response, Tuple[str, int]]

collision_objects: Dict[str, object_type.Models] = {}


@app.route("/collisions/box", methods=['PUT'])
def put_box() -> RespT:
    """Add or update collision box.
        ---
        put:
            tags:
                - Collisions
            description: Add or update collision box.
            parameters:
                - name: boxId
                  in: query
                  schema:
                    type: string
                - name: sizeX
                  in: query
                  schema:
                    type: number
                    format: float
                - name: sizeY
                  in: query
                  schema:
                    type: number
                    format: float
                - name: sizeZ
                  in: query
                  schema:
                    type: number
                    format: float
            requestBody:
                  content:
                    application/json:
                      schema:
                        $ref: Pose
            responses:
                200:
                  description: Ok
    """

    args = convert_keys(request.args.to_dict(), camel_case_to_snake_case)
    box = object_type.Box(args["box_id"], float(args["size_x"]), float(args["size_y"]), float(args["size_z"]))
    collision_objects[box.id] = box
    return "ok", 200


@app.route("/collisions/sphere", methods=['PUT'])
def put_sphere() -> RespT:
    """Add or update collision sphere.
        ---
        put:
            tags:
                - Collisions
            description: Add or update collision sphere.
            parameters:
                - name: sphereId
                  in: query
                  schema:
                    type: string
                - name: radius
                  in: query
                  schema:
                    type: number
                    format: float
            requestBody:
                  content:
                    application/json:
                      schema:
                        $ref: Pose
            responses:
                200:
                  description: Ok
    """

    args = convert_keys(request.args.to_dict(), camel_case_to_snake_case)
    sphere = object_type.Sphere(args["sphere_id"], float(args["radius"]))
    collision_objects[sphere.id] = sphere
    return "ok", 200


@app.route("/collisions/cylinder", methods=['PUT'])
def put_cylinder() -> RespT:
    """Add or update collision cylinder.
        ---
        put:
            tags:
                - Collisions
            description: Add or update collision cylinder.
            parameters:
                - name: cylinderId
                  in: query
                  schema:
                    type: string
                - name: radius
                  in: query
                  schema:
                    type: number
                    format: float
                - name: height
                  in: query
                  schema:
                    type: number
                    format: float
            requestBody:
                  content:
                    application/json:
                      schema:
                        $ref: Pose
            responses:
                200:
                  description: Ok
    """

    args = convert_keys(request.args.to_dict(), camel_case_to_snake_case)
    cylinder = object_type.Cylinder(args["cylinder_id"], float(args["radius"]), float(args["height"]))
    collision_objects[cylinder.id] = cylinder
    return "ok", 200


@app.route("/collisions/mesh", methods=['PUT'])
def put_mesh() -> RespT:
    """Add or update collision mesh.
        ---
        put:
            tags:
                - Collisions
            description: Add or update collision mesh.
            parameters:
                - name: meshId
                  in: query
                  schema:
                    type: string
                - name: uri
                  in: query
                  schema:
                    type: string
                - name: meshScaleX
                  in: query
                  schema:
                    type: number
                    format: float
                    default: 1
                - name: meshScaleY
                  in: query
                  schema:
                    type: number
                    format: float
                    default: 1
                - name: meshScaleZ
                  in: query
                  schema:
                    type: number
                    format: float
                    default: 1
            requestBody:
                  content:
                    application/json:
                      schema:
                        $ref: Pose
            responses:
                200:
                  description: Ok
    """

    args = convert_keys(request.args.to_dict(), camel_case_to_snake_case)
    mesh = object_type.Mesh(args["mesh_id"], args["uri"])
    collision_objects[mesh.id] = mesh
    return "ok", 200


@app.route("/collisions/<string:collisionId>", methods=['DELETE'])
def delete_collision(collisionId: str) -> RespT:
    """Deletes collision object.
        ---
        delete:
            tags:
                - Collisions
            summary: Deletes collision object.
            parameters:
                - name: collisionId
                  in: path
                  description: unique ID
                  required: true
                  schema:
                    type: string
            responses:
                200:
                  description: Ok
    """

    try:
        del collision_objects[collisionId]
    except KeyError:
        return "Not found", 404

    return "ok", 200


@app.route("/collisions", methods=['GET'])
def get_collisions() -> RespT:
    """Gets collision ids.
        ---
        get:
            tags:
            - Collisions
            summary: Gets collision ids.
            responses:
                '200':
                  description: Success
                  content:
                    application/json:
                      schema:
                        type: array
                        items:
                            type: string
    """

    return cast(Response, jsonify(list(collision_objects.keys())))


@app.route("/utils/focus", methods=['PUT'])
def put_focus() -> RespT:
    """Calculates position of object.
        ---
        put:
            tags:
                - Utils
            description: Calculates position of object.
            requestBody:
                  content:
                    application/json:
                      schema:
                        $ref: MeshFocusAction
            responses:
                200:
                  description: Ok
                  content:
                      application/json:
                          schema:
                            $ref: Pose
    """

    return cast(Response, jsonify(common.Pose().to_json()))


@app.route("/swagger/api/swagger.json", methods=["GET"])
def get_swagger() -> str:
    return json.dumps(spec.to_dict())


spec.components.schema(common.Pose.__name__, schema=common.Pose)
spec.components.schema(object_type.Box.__name__, schema=object_type.Box)
spec.components.schema(object_type.Cylinder.__name__, schema=object_type.Cylinder)
spec.components.schema(object_type.Sphere.__name__, schema=object_type.Sphere)
spec.components.schema(object_type.Mesh.__name__, schema=object_type.Mesh)
spec.components.schema(object_type.MeshFocusAction.__name__, schema=object_type.MeshFocusAction)


with app.test_request_context():

    spec.path(view=put_sphere)
    spec.path(view=put_box)
    spec.path(view=put_focus)
    spec.path(view=put_mesh)
    spec.path(view=put_cylinder)
    spec.path(view=delete_collision)
    spec.path(view=get_collisions)


def main() -> None:

    parser = argparse.ArgumentParser(description=SERVICE_NAME)
    parser.add_argument('-s', '--swagger', action="store_true", default=False)
    args = parser.parse_args()

    if args.swagger:
        print(spec.to_yaml())
        return

    SWAGGER_URL = "/swagger"

    swaggerui_blueprint = get_swaggerui_blueprint(
        SWAGGER_URL,  # Swagger UI static files will be mapped to '{SWAGGER_URL}/dist/'
        "./api/swagger.json"
    )

    # Register blueprint at URL
    app.register_blueprint(swaggerui_blueprint, url_prefix=SWAGGER_URL)

    app.run(host='0.0.0.0', port=PORT)


if __name__ == '__main__':
    main()



import re
from dataclasses import dataclass, field
from enum import Enum
from typing import List, NamedTuple, Optional, Set

from dataclasses_jsonschema import JsonSchemaMixin

"""
mypy does not recognize __qualname__ so far: https://github.com/python/mypy/issues/6473
flake8 sucks here as well: https://bugs.launchpad.net/pyflakes/+bug/1648651
TODO: remove type: ignore once it is fixed
"""


def wo_suffix(name: str) -> str:
    return re.sub('Request$', '', name)


class ChangeType(Enum):

    ADD: str = "add"
    UPDATE: str = "update"
    DELETE: str = "delete"


@dataclass
class Request(JsonSchemaMixin):

    id: int
    request: str


@dataclass
class Response(JsonSchemaMixin):

    response: str = ""
    id: int = 0
    result: bool = True
    messages: Optional[List[str]] = None


# ----------------------------------------------------------------------------------------------------------------------

@dataclass
class TypeArgs(JsonSchemaMixin):
    type: str


@dataclass
class IdArgs(JsonSchemaMixin):
    id: str


@dataclass
class RobotArg(JsonSchemaMixin):

    class RobotArgTuple(NamedTuple):

        robot_id: str
        end_effector: str

    robot_id: str = field(metadata=dict(description="Object id of the robot or robot_id within the robot service."))
    end_effector: str

    def as_tuple(self) -> RobotArgTuple:
        return RobotArg.RobotArgTuple(self.robot_id, self.end_effector)


# ----------------------------------------------------------------------------------------------------------------------

@dataclass
class SystemInfoRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class SystemInfoData(JsonSchemaMixin):

    version: str = ""
    api_version: str = ""
    supported_parameter_types: Set[str] = field(default_factory=set)
    supported_rpc_requests: Set[str] = field(default_factory=set)


@dataclass
class SystemInfoResponse(Response):

    data: SystemInfoData = field(default_factory=SystemInfoData)
    response: str = field(default=SystemInfoRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class VersionRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class VersionData(JsonSchemaMixin):

    version: str = ""


@dataclass
class VersionResponse(Response):

    data: VersionData = field(default_factory=VersionData)
    response: str = field(default=VersionRequest.request, init=False)



import asyncio
import base64
import os
import tempfile
import uuid
from typing import Dict, Optional, TYPE_CHECKING

import websockets
from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import helpers as hlp, rest
from arcor2.data import common, events, rpc
from arcor2.exceptions import Arcor2Exception
from arcor2.server import events as server_events, globals as glob, notifications as notif, project

if TYPE_CHECKING:
    ReqQueue = asyncio.Queue[rpc.common.Request]
    RespQueue = asyncio.Queue[rpc.common.Response]
else:
    ReqQueue = asyncio.Queue
    RespQueue = asyncio.Queue

MANAGER_RPC_REQUEST_QUEUE: ReqQueue = ReqQueue()
MANAGER_RPC_RESPONSES: Dict[int, RespQueue] = {}


async def run_temp_package(package_id: str) -> None:

    assert glob.PROJECT
    project_id = glob.PROJECT.id
    glob.TEMPORARY_PACKAGE = True

    await project.close_project(do_cleanup=False)

    exe_req = rpc.execution.RunPackageRequest(uuid.uuid4().int,
                                              args=rpc.execution.RunPackageArgs(package_id, cleanup_after_run=False))
    exe_resp = await manager_request(exe_req)

    if not exe_resp.result:
        glob.logger.warning(f"Execution of temporary package failed with: {exe_resp.messages}.")
    else:
        await server_events.package_started.wait()
        await server_events.package_stopped.wait()
        glob.logger.info("Temporary package stopped, let's remove it and reopen project.")

    glob.TEMPORARY_PACKAGE = False

    await manager_request(rpc.execution.DeletePackageRequest(uuid.uuid4().int, args=rpc.common.IdArgs(package_id)))

    await project.open_project(project_id)

    assert glob.SCENE
    assert glob.PROJECT

    asyncio.ensure_future(notif.broadcast_event(events.OpenProject(data=events.OpenProjectData(glob.SCENE.scene,
                                                                                               glob.PROJECT.project))))


async def build_and_upload_package(project_id: str, package_name: str) -> str:
    """
    Builds package and uploads it to the Execution unit.
    :param project_id:
    :param package_name:
    :return: generated package ID.
    """

    package_id = common.uid()

    # call build service
    # TODO store data in memory
    with tempfile.TemporaryDirectory() as tmpdirname:
        path = os.path.join(tmpdirname, "publish.zip")

        await hlp.run_in_executor(rest.download, f"{glob.BUILDER_URL}/project/{project_id}/publish", path,
                                  None, {"package_name": package_name})

        with open(path, "rb") as zip_file:
            b64_bytes = base64.b64encode(zip_file.read())
            b64_str = b64_bytes.decode()

    # send data to execution service
    exe_req = rpc.execution.UploadPackageRequest(uuid.uuid4().int,
                                                 args=rpc.execution.UploadPackageArgs(package_id, b64_str))
    exe_resp = await manager_request(exe_req)

    if not exe_resp.result:
        if not exe_resp.messages:
            raise Arcor2Exception("Upload to the Execution unit failed.")
        raise Arcor2Exception("\n".join(exe_resp.messages))

    return package_id


async def manager_request(req: rpc.common.Request, ui: Optional[WsClient] = None) -> rpc.common.Response:

    assert req.id not in MANAGER_RPC_RESPONSES

    MANAGER_RPC_RESPONSES[req.id] = RespQueue(maxsize=1)
    await MANAGER_RPC_REQUEST_QUEUE.put(req)

    resp = await MANAGER_RPC_RESPONSES[req.id].get()
    del MANAGER_RPC_RESPONSES[req.id]
    return resp


async def project_manager_client(handle_manager_incoming_messages) -> None:

    while True:

        glob.logger.info("Attempting connection to manager...")

        try:

            async with websockets.connect(glob.MANAGER_URL) as manager_client:

                glob.logger.info("Connected to manager.")

                future = asyncio.ensure_future(handle_manager_incoming_messages(manager_client))

                while True:

                    if future.done():
                        break

                    try:
                        msg = await asyncio.wait_for(MANAGER_RPC_REQUEST_QUEUE.get(), 1.0)
                    except asyncio.TimeoutError:
                        continue

                    try:
                        await manager_client.send(msg.to_json())
                    except websockets.exceptions.ConnectionClosed:
                        await MANAGER_RPC_REQUEST_QUEUE.put(msg)
                        break
        except ConnectionRefusedError as e:
            glob.logger.error(e)
            await asyncio.sleep(delay=1.0)



import copy
import inspect
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Iterator, Optional, Set, Tuple, Type, get_type_hints

import horast

from typed_ast.ast3 import AST, Name

import typing_inspect  # type: ignore

import arcor2
from arcor2.data.common import ActionMetadata
from arcor2.data.object_type import ActionParameterMeta, ObjectAction, ObjectTypeMeta
from arcor2.data.robot import RobotMeta
from arcor2.docstring import parse_docstring
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import Generic, GenericWithPose
from arcor2.parameter_plugins.base import ParameterPlugin, ParameterPluginException
from arcor2.source.utils import SourceException, find_class_def, find_function


class ObjectTypeException(Arcor2Exception):
    pass


@dataclass
class ObjectTypeData:

    meta: ObjectTypeMeta
    type_def: Optional[Type[Generic]] = None
    actions: Dict[str, ObjectAction] = field(default_factory=dict)
    ast: Optional[AST] = None
    robot_meta: Optional[RobotMeta] = None

    def __post_init__(self):
        if not self.meta.disabled:
            assert self.type_def is not None
            assert self.ast is not None


ObjectTypeDict = Dict[str, ObjectTypeData]


def built_in_types() -> Iterator[Tuple[str, Type[Generic]]]:
    """
    Yields class name and class definition tuple
    """

    for cls in inspect.getmembers(arcor2.object_types.abstract, inspect.isclass):
        if not issubclass(cls[1], Generic):
            continue

        yield cls[0], cls[1]


def get_built_in_type(name: str) -> Type[Generic]:

    for bname, cls in built_in_types():
        if name == bname:
            return cls

    raise KeyError


def built_in_types_names() -> Set[str]:

    names = set()

    for type_name, _ in built_in_types():
        names.add(type_name)

    return names


class DataError(Arcor2Exception):
    pass


def obj_description_from_base(data: ObjectTypeDict, obj_type: ObjectTypeMeta) -> str:

    try:
        obj = data[obj_type.base]
    except KeyError:
        raise DataError(f"Unknown object type: {obj_type}.")

    if obj.meta.description:
        return obj.meta.description

    if not obj.meta.base:
        return ""

    return obj_description_from_base(data, data[obj.meta.base].meta)


def meta_from_def(type_def: Type[Generic], built_in: bool = False) -> ObjectTypeMeta:

    obj = ObjectTypeMeta(type_def.__name__,
                         type_def.description(),
                         built_in=built_in,
                         abstract=type_def.abstract(),
                         has_pose=issubclass(type_def, GenericWithPose))

    for base in inspect.getmro(type_def)[1:-1]:  # skip type itself (first base) and the last one (object)
        if issubclass(base, Generic):
            obj.base = base.__name__
            break

    return obj


def base_from_source(source: str, cls_name: str) -> Optional[str]:

    cls_def = find_class_def(cls_name, horast.parse(source))
    if not cls_def.bases:
        return None

    base_name = cls_def.bases[-1]  # allow usage of mixins e.g. class MyType(mixin, Generic)

    assert isinstance(base_name, Name)
    return base_name.id


def built_in_types_data(plugins: Dict[Type, Type[ParameterPlugin]]) -> ObjectTypeDict:

    ret: ObjectTypeDict = {}

    # built-in object types / services
    for _, type_def in built_in_types():

        assert issubclass(type_def, Generic)

        ast = horast.parse(inspect.getsource(type_def))

        d = ObjectTypeData(meta_from_def(type_def, built_in=True),
                           type_def,
                           object_actions(plugins, type_def, ast),
                           ast
                           )

        ret[d.meta.type] = d

    return ret


class IgnoreActionException(Arcor2Exception):
    pass


def _resolve_param(plugins: Dict[Type, Type[ParameterPlugin]], name: str, ttype) -> Type[ParameterPlugin]:

    try:
        return plugins[ttype]
    except KeyError:
        for k, v in plugins.items():
            if not v.EXACT_TYPE and inspect.isclass(ttype) and issubclass(ttype, k):
                return v

    # ignore action with unknown parameter type
    raise IgnoreActionException(f"Parameter {name} has unknown type {ttype}.")


def iterate_over_actions(type_def: Type[Generic]) -> Iterator[Tuple[str, Callable[[Any, ], Any]]]:

    for method_name, method in inspect.getmembers(type_def, inspect.isroutine):

        try:
            if not isinstance(method.__action__, ActionMetadata):
                continue
        except AttributeError:
            continue

        yield method_name, method


def object_actions(plugins: Dict[Type, Type[ParameterPlugin]], type_def: Type[Generic], tree: AST) \
        -> Dict[str, ObjectAction]:

    ret: Dict[str, ObjectAction] = {}

    # ...inspect.ismethod does not work on un-initialized classes
    for method_name, method_def in iterate_over_actions(type_def):

        meta: ActionMetadata = method_def.__action__  # type: ignore

        data = ObjectAction(name=method_name, meta=meta)

        if method_name in type_def.CANCEL_MAPPING:
            meta.cancellable = True

        try:

            if not method_def.__doc__:
                doc = {}
            else:
                doc = parse_docstring(method_def.__doc__)
                doc_short = doc["short_description"]
                if doc_short:
                    data.description = doc_short

            signature = inspect.signature(method_def)

            try:
                method_tree = find_function(method_name, tree)
            except SourceException:
                # function is probably defined in predecessor, will be added later
                continue

            for name, ttype in get_type_hints(method_def).items():

                if name == "return":

                    # ...just ignore NoneType for returns
                    if ttype == type(None):  # noqa: E721
                        continue

                    if typing_inspect.is_tuple_type(ttype):
                        for arg in typing_inspect.get_args(ttype):
                            resolved_param = _resolve_param(plugins, name, arg)
                            if resolved_param is None:
                                raise IgnoreActionException("None in return tuple is not supported.")
                            data.returns.append(resolved_param.type_name())
                    else:
                        # TODO resolving needed for e.g. enums - add possible values to action metadata somewhere?
                        data.returns = [_resolve_param(plugins, name, ttype).type_name()]

                    continue

                param_type = _resolve_param(plugins, name, ttype)

                assert param_type is not None

                args = ActionParameterMeta(name=name, type=param_type.type_name())
                try:
                    param_type.meta(args, method_def, method_tree)
                except ParameterPluginException as e:
                    raise IgnoreActionException(e) from e

                if name in type_def.DYNAMIC_PARAMS:
                    args.dynamic_value = True
                    dvp = type_def.DYNAMIC_PARAMS[name][1]
                    if dvp:
                        args.dynamic_value_parents = dvp

                def_val = signature.parameters[name].default
                if def_val is not inspect.Parameter.empty:
                    args.default_value = def_val

                try:
                    args.description = doc["params"][name].strip()
                except KeyError:
                    pass

                data.parameters.append(args)

        except IgnoreActionException as e:
            data.disabled = True
            data.problem = e.message
            # TODO log exception

        ret[data.name] = data

    return ret


def add_ancestor_actions(obj_type_name: str, object_types: ObjectTypeDict) -> None:

    base_name = object_types[obj_type_name].meta.base

    if not base_name:
        return

    if object_types[base_name].meta.base:
        add_ancestor_actions(base_name, object_types)

    # do not add action from base if it is overridden in child
    # TODO rewrite (use adv. of dict!)
    for base_action in object_types[base_name].actions.values():
        for obj_action in object_types[obj_type_name].actions.values():
            if base_action.name == obj_action.name:

                # built-in object has no "origins" yet
                if not obj_action.origins:
                    obj_action.origins = base_name
                break
        else:
            action = copy.deepcopy(base_action)
            if not action.origins:
                action.origins = base_name
            object_types[obj_type_name].actions[action.name] = action


__all__ = [
    ObjectTypeException.__name__,
    ObjectTypeData.__name__,
    "ObjectTypeDict",
    built_in_types.__name__,
    get_built_in_type.__name__,
    built_in_types_names.__name__,
    DataError.__name__,
    meta_from_def.__name__,
    built_in_types_data.__name__,
    object_actions.__name__,
    add_ancestor_actions.__name__
]



import importlib
import os
import re
import stat
from typing import Any, Dict, List, Optional, Type, Union

import autopep8  # type: ignore

from horast import parse, unparse

from typed_ast.ast3 import AST, Assert, Assign, Attribute, Call, ClassDef, Compare, Eq, ExceptHandler, Expr, \
    FunctionDef, If, ImportFrom, Index, Load, Module, Name, NameConstant, NodeTransformer, NodeVisitor, Pass, Raise, \
    Return, Store, Str, Subscript, Try, While, With, alias, arg, arguments, fix_missing_locations, keyword, stmt, \
    withitem

import typed_astunparse

import arcor2.data.common
from arcor2.cached import CachedProject
from arcor2.data.common import ActionPoint
from arcor2.source import SCRIPT_HEADER, SourceException


def main_loop_body(tree: Module) -> List[Any]:
    main = find_function("main", tree)

    for node in main.body:
        if isinstance(node, While):  # TODO more specific condition (test for True argument)
            return node.body

    raise SourceException("Main loop not found.")


def empty_script_tree(add_main_loop: bool = True) -> Module:
    """
    Creates barebones of the script (empty 'main' function).

    Returns
    -------

    """

    main_body: List[stmt] = []

    if add_main_loop:
        main_body.append(While(
            test=NameConstant(value=True),
            body=[Pass()],
            orelse=[]))
    else:
        """
        put there "pass" in order to make code valid even if there is no other statement (e.g. no object from resources)
        """
        main_body.append(Pass())

    # TODO helper function for try ... except

    tree = Module(
        body=[
            FunctionDef(
                name='main',
                args=arguments(
                    args=[arg(
                        arg='res',
                        annotation=Name(
                            id='Resources',
                            ctx=Load()),
                        type_comment=None)],
                    vararg=None,
                    kwonlyargs=[],
                    kw_defaults=[],
                    kwarg=None,
                    defaults=[]),
                body=main_body,
                decorator_list=[],
                returns=NameConstant(value=None),
                type_comment=None),
            If(
                test=Compare(
                    left=Name(
                        id='__name__',
                        ctx=Load()),
                    ops=[Eq()],
                    comparators=[Str(
                        s='__main__',
                        kind='')]),
                body=[Try(
                    body=[With(
                        items=[withitem(
                            context_expr=Call(
                                func=Name(
                                    id='Resources',
                                    ctx=Load()),
                                args=[],
                                keywords=[]),
                            optional_vars=Name(
                                id='res',
                                ctx=Store()))],
                        body=[Expr(value=Call(
                            func=Name(
                                id='main',
                                ctx=Load()),
                            args=[Name(
                                id='res',
                                ctx=Load())],
                            keywords=[]))],
                        type_comment=None)],
                    handlers=[ExceptHandler(
                        type=Name(
                            id='Exception',
                            ctx=Load()),
                        name='e',
                        body=[Expr(value=Call(
                            func=Name(
                                id='print_exception',
                                ctx=Load()),
                            args=[Name(
                                id='e',
                                ctx=Load())],
                            keywords=[]))])],
                    orelse=[],
                    finalbody=[])],
                orelse=[])],
        type_ignores=[])

    add_import(tree, "arcor2.helpers", "print_exception")
    add_import(tree, "resources", "Resources", try_to_import=False)

    return tree


def find_asserts(tree: FunctionDef) -> List[Assert]:
    class FindAsserts(NodeVisitor):

        def __init__(self) -> None:
            self.asserts: List[Assert] = []

        def visit_Assert(self, node: Assert) -> None:
            self.asserts.append(node)

    ff = FindAsserts()
    ff.visit(tree)

    return ff.asserts


def find_function(name: str, tree: Union[Module, AST]) -> FunctionDef:
    class FindFunction(NodeVisitor):

        def __init__(self) -> None:
            self.function_node: Optional[FunctionDef] = None

        def visit_FunctionDef(self, node: FunctionDef) -> None:
            if node.name == name:
                self.function_node = node
                return

            if not self.function_node:
                self.generic_visit(node)

    ff = FindFunction()
    ff.visit(tree)

    if ff.function_node is None:
        raise SourceException(f"Function {name} not found.")

    return ff.function_node


def find_class_def(name: str, tree: Union[Module, AST]) -> ClassDef:
    class FindClassDef(NodeVisitor):

        def __init__(self) -> None:
            self.cls_def_node: Optional[ClassDef] = None

        def visit_ClassDef(self, node: ClassDef) -> None:
            if node.name == name:
                self.cls_def_node = node
                return

            if not self.cls_def_node:
                self.generic_visit(node)

    ff = FindClassDef()
    ff.visit(tree)

    if ff.cls_def_node is None:
        raise SourceException(f"Class definition {name} not found.")

    return ff.cls_def_node


def add_import(node: Module, module: str, cls: str, try_to_import: bool = True) -> None:
    """
    Adds "from ... import ..." to the beginning of the script.

    Parameters
    ----------
    node
    module
    cls

    Returns
    -------

    """

    class AddImportTransformer(NodeTransformer):

        def __init__(self, module: str, cls: str) -> None:
            self.done = False
            self.module = module
            self.cls = cls

        def visit_ImportFrom(self, node: ImportFrom) -> ImportFrom:
            if node.module == self.module:

                for aliass in node.names:
                    if aliass.name == self.cls:
                        self.done = True
                        break
                else:
                    node.names.append(alias(name=self.cls, asname=None))
                    self.done = True

            return node

    if try_to_import:

        try:
            imported_mod = importlib.import_module(module)
        except ModuleNotFoundError as e:
            raise SourceException(e)

        try:
            getattr(imported_mod, cls)
        except AttributeError as e:
            raise SourceException(e)

    tr = AddImportTransformer(module, cls)
    node = tr.visit(node)

    if not tr.done:
        node.body.insert(0, ImportFrom(module=module, names=[alias(name=cls, asname=None)], level=0))


def add_cls_inst(node: Module, cls: str, name: str, kwargs: Optional[Dict] = None,
                 kwargs2parse: Optional[Dict] = None) -> None:
    class FindImport(NodeVisitor):

        def __init__(self, cls: str) -> None:

            self.found = False
            self.cls = cls

        def visit_ImportFrom(self, node: ImportFrom) -> None:

            for node_alias in node.names:
                if node_alias.name == self.cls:
                    self.found = True

            if not self.found:
                self.generic_visit(node)

    class FindClsInst(NodeVisitor):

        def __init__(self) -> None:

            self.found = False

        def visit_FunctionDef(self, node: FunctionDef) -> None:

            if node.name == 'main':

                for item in node.body:

                    if isinstance(item, Assign):

                        assert isinstance(item.targets[0], Name)

                        if item.targets[0].id == name:

                            # TODO assert for item.value
                            if item.value.func.id != cls:  # type: ignore
                                raise SourceException(
                                    "Name '{}' already used for instance of '{}'!".
                                    format(name, item.value.func.id))  # type: ignore

                            self.found = True
                            # TODO update arguments?

            if not self.found:
                self.generic_visit(node)

    class AddClsInst(NodeTransformer):

        def visit_FunctionDef(self, node: FunctionDef) -> FunctionDef:

            if node.name == 'main':

                kw = []

                if kwargs:
                    for k, v in kwargs.items():
                        kw.append(keyword(arg=k, value=v))

                if kwargs2parse:
                    for k, v in kwargs2parse.items():
                        kw.append(keyword(arg=k, value=parse(v)))

                node.body.insert(0, Assign(targets=[Name(id=name, ctx=Store())],
                                           value=Call(func=Name(id=cls, ctx=Load()),
                                                      args=[],
                                                      keywords=kw)))

            return node

    find_import = FindImport(cls)
    find_import.visit(node)

    if not find_import.found:
        raise SourceException("Class '{}' not imported!".format(cls))

    vis = FindClsInst()
    vis.visit(node)

    if not vis.found:
        tr = AddClsInst()
        node = tr.visit(node)


def append_method_call(body: List, instance: str, method: str, args: List, kwargs: List) -> None:
    body.append(Expr(value=Call(func=get_name_attr(instance, method),
                                args=args,
                                keywords=kwargs)))


def add_method_call_in_main(tree: Module, instance: str, method: str, args: List, kwargs: List) -> None:
    """
    Places method call after block where instances are created.

    Parameters
    ----------
    tree
    instance
    init
    args
    kwargs

    Returns
    -------

    """

    main_body = find_function("main", tree).body
    last_assign_idx = None

    for body_idx, body_item in enumerate(main_body):

        # TODO check if instance exists!

        if isinstance(body_item, Assign):
            last_assign_idx = body_idx

    if not last_assign_idx:
        raise SourceException()

    # TODO iterate over args/kwargs
    # TODO check actual number of method's arguments (and types?)
    main_body.insert(last_assign_idx + 1, Expr(value=Call(func=get_name_attr(instance, method),
                                                          args=args, keywords=[])))


def get_name(name: str) -> Name:
    return Name(id=name, ctx=Load())


def get_name_attr(name: str, attr: str, ctx: Union[Type[Load], Type[Store]] = Load) -> Attribute:
    return Attribute(
        value=get_name(name),
        attr=attr,
        ctx=ctx())


def tree_to_str(tree: AST) -> str:
    # TODO why this fails?
    # validator.visit(tree)

    fix_missing_locations(tree)
    generated_code: str = unparse(tree)
    generated_code = autopep8.fix_code(generated_code, options={'aggressive': 1})

    return generated_code


def make_executable(path_to_file: str) -> None:
    st = os.stat(path_to_file)
    os.chmod(path_to_file, st.st_mode | stat.S_IEXEC)


def tree_to_script(tree: Module, out_file: str, executable: bool) -> None:
    generated_code = tree_to_str(tree)

    with open(out_file, "w") as f:
        f.write(SCRIPT_HEADER)
        f.write(generated_code)

    if executable:
        make_executable(out_file)


def clean(x: str) -> str:
    return re.sub('\W|^(?=\d)', '_', x)  # noqa


def global_action_points_class(project: CachedProject) -> str:
    tree = Module(body=[])
    tree.body.append(ImportFrom(
        module=arcor2.data.common.__name__,
        names=[alias(
            name=ActionPoint.__name__,
            asname=None)],
        level=0))
    tree.body.append(ImportFrom(
        module='resources',
        names=[alias(
            name='Resources',
            asname=None)],
        level=0))

    cls_def = ClassDef(
        name='ActionPoints',
        bases=[],
        keywords=[],
        body=[
            FunctionDef(
                name='__init__',
                args=arguments(
                    args=[
                        arg(
                            arg='self',
                            annotation=None,
                            type_comment=None),
                        arg(
                            arg='res',
                            annotation=Name(
                                id='Resources',
                                ctx=Load()),
                            type_comment=None)],
                    vararg=None,
                    kwonlyargs=[],
                    kw_defaults=[],
                    kwarg=None,
                    defaults=[]),
                body=[Assign(
                    targets=[Attribute(
                        value=Name(
                            id='self',
                            ctx=Load()),
                        attr='_res',
                        ctx=Store())],
                    value=Name(
                        id='res',
                        ctx=Load()),
                    type_comment=None)],
                decorator_list=[],
                returns=None,
                type_comment=None)],
        decorator_list=[])

    for ap in project.action_points:
        fd = FunctionDef(
            name=clean(ap.name),  # TODO avoid possible collisions
            args=arguments(
                args=[arg(arg='self', annotation=None, type_comment=None)],
                vararg=None,
                kwonlyargs=[],
                kw_defaults=[],
                kwarg=None,
                defaults=[]),
            body=[Return(value=Call(
                func=Attribute(
                    value=Attribute(
                        value=Attribute(
                            value=Name(
                                id='self',
                                ctx=Load()),
                            attr='_res',
                            ctx=Load()),
                        attr='project',
                        ctx=Load()),
                    attr='action_point',
                    ctx=Load()),
                args=[Str(
                    s=ap.id,
                    kind='')],
                keywords=[]))],
            decorator_list=[Name(
                id='property',
                ctx=Load())],
            returns=Name(
                id='ActionPoint',
                ctx=Load()),
            type_comment=None)

        cls_def.body.append(fd)

    tree.body.append(cls_def)
    return tree_to_str(tree)


def global_actions_class(project: CachedProject) -> str:
    tree = Module(body=[])
    tree.body.append(ImportFrom(
        module='resources',
        names=[alias(
            name='Resources',
            asname=None)],
        level=0))

    cls_def = ClassDef(
        name='Actions',
        bases=[],
        keywords=[],
        body=[
            FunctionDef(
                name='__init__',
                args=arguments(
                    args=[
                        arg(
                            arg='self',
                            annotation=None,
                            type_comment=None),
                        arg(
                            arg='res',
                            annotation=Name(
                                id='Resources',
                                ctx=Load()),
                            type_comment=None)],
                    vararg=None,
                    kwonlyargs=[],
                    kw_defaults=[],
                    kwarg=None,
                    defaults=[]),
                body=[Assign(
                    targets=[Attribute(
                        value=Name(
                            id='self',
                            ctx=Load()),
                        attr='_res',
                        ctx=Store())],
                    value=Name(
                        id='res',
                        ctx=Load()),
                    type_comment=None)],
                decorator_list=[],
                returns=None,
                type_comment=None)],
        decorator_list=[])

    for action in project.actions:

        ac_obj, ac_type = action.parse_type()

        m = FunctionDef(
            name=clean(action.name),
            args=arguments(
                args=[arg(
                    arg='self',
                    annotation=None,
                    type_comment=None)],
                vararg=None,
                kwonlyargs=[],
                kw_defaults=[],
                kwarg=None,
                defaults=[]),
            body=[Expr(value=Call(
                func=Attribute(
                    value=Subscript(
                        value=Attribute(
                            value=Attribute(
                                value=Name(
                                    id='self',
                                    ctx=Load()),
                                attr='_res',
                                ctx=Load()),
                            attr='all_instances',
                            ctx=Load()),
                        slice=Index(value=Str(
                            s=ac_obj,
                            kind='')),
                        ctx=Load()),
                    attr=ac_type,
                    ctx=Load()),
                args=[Attribute(
                    value=Attribute(
                        value=Name(
                            id='self',
                            ctx=Load()),
                        attr='_res',
                        ctx=Load()),
                    attr=clean(action.name),
                    ctx=Load())],
                keywords=[]))],
            decorator_list=[],
            returns=None,
            type_comment=None)

        cls_def.body.append(m)

    tree.body.append(cls_def)
    return tree_to_str(tree)


def derived_resources_class(project: CachedProject) -> str:
    # TODO temporary and ugly solution of circular import
    import arcor2.resources
    from arcor2.resources import ResourcesBase

    tree = Module(body=[])

    parameters = [(act.id, clean(act.name)) for act in project.actions]

    add_import(tree, arcor2.resources.__name__, ResourcesBase.__name__)

    derived_cls_name = "Resources"

    init_body: List = [Expr(value=Call(
        func=Attribute(value=Call(func=Name(id='super', ctx=Load()),
                                  args=[Name(id=derived_cls_name, ctx=Load()),
                                        Name(id='self', ctx=Load())], keywords=[]),
                       attr='__init__', ctx=Load()), args=[Str(s=project.id)],
        keywords=[]))]

    for a_id, a_name in parameters:
        init_body.append(Assign(targets=[get_name_attr("self", "_" + a_name, Store)],
                                value=Call(
                                    func=get_name_attr("self", "parameters"),
                                    args=[Str(s=a_id)], keywords=[])))

    cls_def = ClassDef(name=derived_cls_name,
                       bases=[Name(id=ResourcesBase.__name__, ctx=Load())],
                       keywords=[],
                       body=[FunctionDef(name='__init__', args=arguments(args=[arg(arg='self', annotation=None)],
                                                                         vararg=None, kwonlyargs=[],
                                                                         kw_defaults=[], kwarg=None,
                                                                         defaults=[]), body=init_body,
                                         decorator_list=[], returns=None)], decorator_list=[])

    tree.body.append(cls_def)

    for a_id, a_name in parameters:
        cls_def.body.append(FunctionDef(
            name=a_name,
            args=arguments(
                args=[arg(
                    arg='self',
                    annotation=None,
                    type_comment=None)],
                vararg=None,
                kwonlyargs=[],
                kw_defaults=[],
                kwarg=None,
                defaults=[]),
            body=[
                Expr(value=Call(
                    func=Attribute(value=Name(id='self', ctx=Load()), attr='print_info', ctx=Load()),
                    args=[
                        Str(
                            s=a_id,
                            kind=''),
                        get_name_attr('self', '_' + a_name)],
                    keywords=[])),
                Return(value=get_name_attr('self', '_' + a_name))],
            decorator_list=[Name(
                id='property',
                ctx=Load())],
            returns=None,
            type_comment=None))

    return tree_to_str(tree)


def dump(tree: Module) -> str:
    return typed_astunparse.dump(tree)


def find_raises(tree: FunctionDef) -> List[Raise]:
    class FindRaises(NodeVisitor):

        def __init__(self) -> None:
            self.raises: List[Raise] = []

        def visit_Raise(self, node: Raise) -> None:
            self.raises.append(node)

    ff = FindRaises()
    ff.visit(tree)

    return ff.raises


def function_implemented(tree: AST, func_name: str) -> bool:
    """
    Body of unimplemented function (e.g. object/service feature) contains only 'raise NotImplementedError()'
    :param tree:
    :return:
    """

    try:
        func_def = find_function(func_name, tree)
    except SourceException:
        return False

    raises = find_raises(func_def)

    if len(raises) != 1:
        return True

    exc = raises[0].exc

    if isinstance(exc, Call):
        # raise NotImplementedError("something")

        assert isinstance(exc.func, Name)

        if exc.func.id == NotImplementedError.__name__:
            return False

    if isinstance(exc, Name) and exc.id == NotImplementedError.__name__:
        # raise NotImplementedError
        return False

    return True



from arcor2.data.rpc import common
from arcor2.data.rpc import execution
from arcor2.data.rpc import objects
from arcor2.data.rpc import project
from arcor2.data.rpc import robot
from arcor2.data.rpc import scene
from arcor2.data.rpc import storage

__all__ = [
    common.__name__,
    execution.__name__,
    objects.__name__,
    project.__name__,
    robot.__name__,
    scene.__name__,
    storage.__name__
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import uuid
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum, unique
from json import JSONEncoder
from typing import Any, ClassVar, Iterator, List, NamedTuple, Optional, Set, cast

from dataclasses_jsonschema import JsonSchemaMixin

import quaternion  # type: ignore

from arcor2.exceptions import Arcor2Exception


def uid() -> str:
    return uuid.uuid4().hex


@unique
class StrEnum(str, Enum):

    @classmethod
    def set(cls) -> Set[str]:
        return set(map(lambda c: c.value, cls))  # type: ignore


@unique
class IntEnum(int, Enum):

    @classmethod
    def set(cls) -> Set[int]:
        return set(map(lambda c: c.value, cls))  # type: ignore


class FlowTypes(StrEnum):

    DEFAULT: str = "default"


class LinkToActionOutput(NamedTuple):
    action_id: str
    flow_name: FlowTypes
    output_index: int


def parse_link(val: str) -> LinkToActionOutput:

    action_id, flow_name, output_idx_str = val.split("/")

    try:
        return LinkToActionOutput(action_id, FlowTypes(flow_name), int(output_idx_str))
    except ValueError as e:
        raise Arcor2Exception("Invalid link value.") from e


class DataClassEncoder(JSONEncoder):

    def default(self, o: Any) -> Any:

        if isinstance(o, JsonSchemaMixin):
            return o.to_dict()

        return super().default(o)


@dataclass
class IterableIndexable(JsonSchemaMixin):

    def __getitem__(self, item: int) -> float:

        attr = getattr(self, tuple(self.__dict__.keys())[item])
        assert isinstance(attr, float)
        return attr

    def __iter__(self) -> Iterator[float]:

        yield from self.__dict__.values()


@dataclass
class Position(IterableIndexable):

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def rotated(self, rot: "Orientation", inverse: bool = False) -> "Position":

        q = rot.as_quaternion()

        if inverse:
            q = q.inverse()

        rotated_vector = quaternion.rotate_vectors([q], [list(self)])[0][0]
        return Position(rotated_vector[0], rotated_vector[1], rotated_vector[2])


@dataclass
class Orientation(IterableIndexable):

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0
    w: float = 1.0

    def as_quaternion(self) -> quaternion.quaternion:
        return quaternion.quaternion(self.w, self.x, self.y, self.z).normalized()

    def set_from_quaternion(self, q: quaternion.quaternion) -> None:

        nq = q.normalized()

        self.x = nq.x
        self.y = nq.y
        self.z = nq.z
        self.w = nq.w

    def __eq__(self, other: object) -> bool:

        if not isinstance(other, Orientation):
            return False

        return cast(bool, quaternion.isclose(self.as_quaternion(), other.as_quaternion(), rtol=1.e-8)[0])


@dataclass
class NamedOrientation(JsonSchemaMixin):

    id: str
    name: str
    orientation: Orientation


@dataclass
class Pose(JsonSchemaMixin):

    position: Position = field(default_factory=Position)
    orientation: Orientation = field(default_factory=Orientation)


@dataclass
class ActionMetadata(JsonSchemaMixin):

    blocking: bool = False
    composite: bool = False
    blackbox: bool = False
    cancellable: bool = field(init=False, default=False)


@dataclass
class Joint(JsonSchemaMixin):

    name: str
    value: float


@dataclass
class ProjectRobotJoints(JsonSchemaMixin):

    id: str
    name: str
    robot_id: str
    joints: List[Joint]
    is_valid: bool = False


@dataclass
class ActionPoint(JsonSchemaMixin):

    id: str
    name: str
    position: Position
    parent: Optional[str] = None
    orientations: List[NamedOrientation] = field(default_factory=list)
    robot_joints: List[ProjectRobotJoints] = field(default_factory=list)

    def orientation_names(self) -> Set[str]:
        return {ori.name for ori in self.orientations}

    def joints_names(self) -> Set[str]:
        return {joints.name for joints in self.robot_joints}

    def invalidate_joints(self) -> None:

        for joints in self.robot_joints:
            joints.is_valid = False


@dataclass
class SceneObjectSetting(JsonSchemaMixin):

    key: str
    type: str
    value: str


@dataclass
class SceneObject(JsonSchemaMixin):

    id: str
    name: str
    type: str
    pose: Optional[Pose] = None
    settings: List[SceneObjectSetting] = field(default_factory=list)
    children: List["SceneObject"] = field(default_factory=list)


@dataclass
class Scene(JsonSchemaMixin):

    id: str
    name: str
    objects: List[SceneObject] = field(default_factory=list)
    desc: str = field(default_factory=str)
    modified: Optional[datetime] = None
    int_modified: Optional[datetime] = None


@dataclass
class IdValue(JsonSchemaMixin):

    id: str
    value: str


class ActionParameterException(Arcor2Exception):
    pass


@dataclass
class ActionParameter(JsonSchemaMixin):

    class TypeEnum(StrEnum):

        CONSTANT: str = "constant"
        LINK: str = "link"

    id: str
    type: str
    value: str

    def parse_link(self) -> LinkToActionOutput:
        assert self.type == ActionParameter.TypeEnum.LINK
        return parse_link(self.value)

    def is_value(self) -> bool:
        return self.type not in ActionParameter.TypeEnum.set()


@dataclass
class Flow(JsonSchemaMixin):

    type: FlowTypes = FlowTypes.DEFAULT
    outputs: List[str] = field(default_factory=list)  # can't be set as it is unordered

    def __post_init__(self) -> None:

        if len(self.outputs) > len(set(self.outputs)):
            raise Arcor2Exception("Outputs have to be unique.")


@dataclass
class Action(JsonSchemaMixin):

    class ParsedType(NamedTuple):

        obj_id: str
        action_type: str

    id: str
    name: str
    type: str
    parameters: List[ActionParameter] = field(default_factory=list)
    flows: List[Flow] = field(default_factory=list)

    def parse_type(self) -> ParsedType:

        try:
            obj_id_str, action = self.type.split("/")
        except ValueError:
            raise Arcor2Exception(f"Action: {self.id} has invalid type: {self.type}.")
        return Action.ParsedType(obj_id_str, action)

    def parameter(self, parameter_id: str) -> ActionParameter:

        for param in self.parameters:
            if parameter_id == param.id:
                return param

        raise Arcor2Exception("Param not found")

    @property
    def bare(self) -> "Action":
        return Action(self.id, self.name, self.type)

    def flow(self, flow_type: FlowTypes = FlowTypes.DEFAULT) -> Flow:

        for flow in self.flows:
            if flow.type == flow_type:
                return flow
        raise Arcor2Exception(f"Flow '{flow_type.value}' not found.")


@dataclass
class ProjectActionPoint(ActionPoint):

    actions: List[Action] = field(default_factory=list)

    def action_ids(self) -> Set[str]:
        return {action.id for action in self.actions}

    def bare(self) -> "ProjectActionPoint":
        return ProjectActionPoint(self.id, self.name, self.position, self.parent)


@dataclass
class ProjectLogicIf(JsonSchemaMixin):

    what: str
    value: str

    def parse_what(self) -> LinkToActionOutput:
        return parse_link(self.what)


@dataclass
class LogicItem(JsonSchemaMixin):

    class ParsedStart(NamedTuple):

        start_action_id: str
        start_flow: str

    START: ClassVar[str] = "START"
    END: ClassVar[str] = "END"

    id: str
    start: str
    end: str
    condition: Optional[ProjectLogicIf] = None

    def parse_start(self) -> ParsedStart:

        try:
            start_action_id, start_flow = self.start.split("/")
        except ValueError:
            return LogicItem.ParsedStart(self.start, FlowTypes.DEFAULT)

        return LogicItem.ParsedStart(start_action_id, start_flow)


@dataclass
class ProjectConstant(JsonSchemaMixin):

    id: str
    name: str
    type: str
    value: str


@dataclass
class FunctionReturns(JsonSchemaMixin):

    type: str
    link: str


@dataclass
class ProjectFunction(JsonSchemaMixin):

    id: str
    name: str
    actions: List[Action] = field(default_factory=list)
    logic: List[LogicItem] = field(default_factory=list)
    parameters: List[ActionParameter] = field(default_factory=list)
    returns: List[FunctionReturns] = field(default_factory=list)

    def action_ids(self) -> Set[str]:
        return {act.id for act in self.actions}

    def action(self, action_id: str) -> Action:

        for ac in self.actions:
            if ac.id == action_id:
                return ac
        else:
            raise Arcor2Exception("Action not found")


@dataclass
class Project(JsonSchemaMixin):

    id: str
    name: str
    scene_id: str
    desc: str = field(default_factory=str)
    has_logic: bool = True
    modified: Optional[datetime] = None
    int_modified: Optional[datetime] = None
    action_points: List[ProjectActionPoint] = field(default_factory=list)
    constants: List[ProjectConstant] = field(default_factory=list)
    functions: List[ProjectFunction] = field(default_factory=list)
    logic: List[LogicItem] = field(default_factory=list)


@dataclass
class ProjectSources(JsonSchemaMixin):

    id: str  # project_id
    script: str


@dataclass
class IdDesc(JsonSchemaMixin):
    id: str
    name: str
    desc: Optional[str] = None


@dataclass
class IdDescList(JsonSchemaMixin):

    items: List[IdDesc] = field(default_factory=list)


class PackageStateEnum(Enum):

    RUNNING: str = "running"
    STOPPED: str = "stopped"
    PAUSED: str = "paused"
    UNDEFINED: str = "undefined"


class ActionStateEnum(Enum):

    BEFORE: str = "before"
    AFTER: str = "after"


@dataclass
class ActionState(JsonSchemaMixin):

    object_id: str = ""
    method: str = ""
    where: ActionStateEnum = ActionStateEnum.BEFORE


@dataclass
class PackageState(JsonSchemaMixin):

    state: PackageStateEnum = PackageStateEnum.UNDEFINED
    package_id: Optional[str] = None


@dataclass
class CurrentAction(JsonSchemaMixin):

    action_id: str = ""
    args: List[ActionParameter] = field(default_factory=list)


@dataclass
class BroadcastInfo(JsonSchemaMixin):

    host: str
    port: int



from copy import deepcopy
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Dict, Iterator, List, Optional, Set, Tuple, ValuesView

from arcor2.data.common import Action, LogicItem, NamedOrientation, Pose, Position, Project, ProjectActionPoint,\
    ProjectConstant, ProjectFunction, ProjectRobotJoints, Scene, SceneObject
from arcor2.exceptions import Arcor2Exception

# TODO cached ProjectFunction (actions from functions are totally ignored at the moment)


class CachedSceneException(Arcor2Exception):
    pass


class CachedScene:

    def __init__(self, scene: Scene):

        self.id: str = scene.id
        self.name: str = scene.name
        self.desc: str = scene.desc
        self.modified: Optional[datetime] = scene.modified
        self.int_modified: Optional[datetime] = scene.int_modified

        # TODO deal with children
        self._objects: Dict[str, SceneObject] = {}

        for obj in scene.objects:

            if obj.id in self._objects:
                raise CachedSceneException(f"Duplicate object id: {obj.id}.")

            self._objects[obj.id] = obj

    @property
    def bare(self) -> Scene:
        return Scene(self.id, self.name, desc=self.desc)

    def object_names(self) -> Iterator[str]:

        for obj in self._objects.values():
            yield obj.name

    @property
    def objects(self) -> Iterator[SceneObject]:

        for obj in self._objects.values():
            yield obj

    @property
    def object_ids(self) -> Set[str]:
        return set(self._objects.keys())

    def object(self, object_id: str) -> SceneObject:

        try:
            return self._objects[object_id]
        except KeyError:
            raise Arcor2Exception(f"Object ID {object_id} not found.")

    def objects_of_type(self, obj_type: str) -> Iterator[SceneObject]:

        for obj in self.objects:
            if obj.type == obj_type:
                yield obj

    @property
    def object_types(self) -> Set[str]:
        return {obj.type for obj in self.objects}

    @property
    def scene(self) -> Scene:

        sc = self.bare
        sc.modified = self.modified
        sc.int_modified = self.int_modified
        sc.objects = list(self.objects)
        return sc


class UpdateableCachedScene(CachedScene):

    def update_modified(self) -> None:
        self.int_modified = datetime.now(tz=timezone.utc)

    def has_changes(self) -> bool:

        if self.int_modified is None:
            return False

        if self.modified is None:
            return True

        return self.int_modified > self.modified

    def upsert_object(self, obj: SceneObject) -> None:

        self._objects[obj.id] = obj
        self.update_modified()

    def delete_object(self, obj_id: str) -> None:

        try:
            del self._objects[obj_id]
        except KeyError as e:
            raise Arcor2Exception("Object id not found.") from e

        self.update_modified()


class CachedProjectException(Arcor2Exception):
    pass


@dataclass
class Actions:

    data: Dict[str, Action] = field(default_factory=dict)
    parent: Dict[str, ProjectActionPoint] = field(default_factory=dict)


@dataclass
class Joints:

    data: Dict[str, ProjectRobotJoints] = field(default_factory=dict)
    parent: Dict[str, ProjectActionPoint] = field(default_factory=dict)


@dataclass
class Orientations:

    data: Dict[str, NamedOrientation] = field(default_factory=dict)
    parent: Dict[str, ProjectActionPoint] = field(default_factory=dict)


class CachedProject:

    def __init__(self, project: Project):

        project_copy = deepcopy(project)  # do not modify original project

        self.id: str = project_copy.id
        self.name: str = project_copy.name
        self.scene_id: str = project_copy.scene_id
        self.desc: str = project_copy.desc
        self.has_logic: bool = project_copy.has_logic
        self.modified: Optional[datetime] = project_copy.modified
        self._int_modified: Optional[datetime] = project_copy.int_modified

        self._action_points: Dict[str, ProjectActionPoint] = {}

        self._actions = Actions()
        self._joints = Joints()
        self._orientations = Orientations()

        self._constants: Dict[str, ProjectConstant] = {}
        self._logic_items: Dict[str, LogicItem] = {}
        self._functions: Dict[str, ProjectFunction] = {}

        for ap in project_copy.action_points:

            if ap.id in self._action_points:
                raise CachedProjectException(f"Duplicate AP id: {ap.id}.")

            self._action_points[ap.id] = ap

            for ac in ap.actions:

                if ac.id in self._actions.data:
                    raise CachedProjectException(f"Duplicate action id: {ac.id}.")

                self._actions.data[ac.id] = ac
                self._actions.parent[ac.id] = ap

            for joints in ap.robot_joints:

                if joints.id in self._joints.data:
                    raise CachedProjectException(f"Duplicate joints id: {joints.id}.")

                self._joints.data[joints.id] = joints
                self._joints.parent[joints.id] = ap

            for orientation in ap.orientations:

                if orientation.id in self._orientations.data:
                    raise CachedProjectException(f"Duplicate orientation id: {orientation.id}.")

                self._orientations.data[orientation.id] = orientation
                self._orientations.parent[orientation.id] = ap

            ap.actions.clear()
            ap.robot_joints.clear()
            ap.orientations.clear()

        for constant in project_copy.constants:
            self._constants[constant.id] = constant

        for logic_item in project_copy.logic:
            self._logic_items[logic_item.id] = logic_item

        for function in project_copy.functions:
            self._functions[function.id] = function

    @property
    def logic(self) -> ValuesView[LogicItem]:
        return self._logic_items.values()

    @property
    def constants(self) -> ValuesView[ProjectConstant]:
        return self._constants.values()

    @property
    def functions(self) -> ValuesView[ProjectFunction]:
        return self._functions.values()

    @property
    def project(self) -> Project:

        proj = Project(self.id, self.name, self.scene_id, self.desc, self.has_logic, self.modified, self._int_modified)

        proj.action_points = list(self._action_points.values())

        for ap in proj.action_points:

            ap.actions.clear()
            ap.robot_joints.clear()
            ap.orientations.clear()

            for act_id, parent_ap in self._actions.parent.items():
                if ap == parent_ap:
                    ap.actions.append(self._actions.data[act_id])

            for joints_id, parent_ap in self._joints.parent.items():
                if ap == parent_ap:
                    ap.robot_joints.append(self._joints.data[joints_id])

            for ori_id, parent_ap in self._orientations.parent.items():
                if ap == parent_ap:
                    ap.orientations.append(self._orientations.data[ori_id])

        proj.constants = list(self.constants)
        proj.functions = list(self.functions)
        proj.logic = list(self.logic)
        return proj

    @property
    def bare(self) -> Project:
        return Project(self.id, self.name, self.scene_id, self.desc, self.has_logic)

    @property
    def action_points(self) -> ValuesView[ProjectActionPoint]:
        return self._action_points.values()

    @property
    def action_points_with_parent(self) -> List[ProjectActionPoint]:
        """
        Get action points which are relative to something (parent is set).
        :return:
        """

        return [ap for ap in self._action_points.values() if ap.parent]

    @property
    def action_points_names(self) -> Set[str]:
        return {ap.name for ap in self._action_points.values()}

    @property
    def action_points_ids(self) -> Set[str]:
        return set(self._action_points.keys())

    def ap_and_joints(self, joints_id: str) -> Tuple[ProjectActionPoint, ProjectRobotJoints]:

        try:
            return self._joints.parent[joints_id], self._joints.data[joints_id]
        except KeyError:
            raise CachedProjectException("Unknown joints.")

    def joints(self, joints_id: str) -> ProjectRobotJoints:

        try:
            return self._joints.data[joints_id]
        except KeyError:
            raise CachedProjectException("Unknown joints.")

    def ap_and_orientation(self, orientation_id: str) -> Tuple[ProjectActionPoint, NamedOrientation]:

        try:
            return self._orientations.parent[orientation_id], self._orientations.data[orientation_id]
        except KeyError:
            raise CachedProjectException("Unknown orientation.")

    def orientation(self, orientation_id: str) -> NamedOrientation:

        try:
            return self._orientations.data[orientation_id]
        except KeyError:
            raise CachedProjectException("Unknown orientation.")

    def pose(self, orientation_id: str) -> Pose:

        ap, ori = self.ap_and_orientation(orientation_id)
        return Pose(ap.position, ori.orientation)

    def action(self, action_id: str) -> Action:

        try:
            return self._actions.data[action_id]
        except KeyError:
            raise CachedProjectException("Action not found")

    def action_point_and_action(self, action_id: str) -> Tuple[ProjectActionPoint, Action]:

        try:
            return self._actions.parent[action_id], self._actions.data[action_id]
        except KeyError:
            raise CachedProjectException("Action not found")

    @property
    def actions(self) -> List[Action]:
        return list(self._actions.data.values())

    def action_ids(self) -> Set[str]:
        return {action.id for action in self.actions}

    def action_user_names(self) -> Set[str]:
        return {action.name for action in self.actions}

    def action_point(self, action_point_id: str) -> ProjectActionPoint:

        try:
            return self._action_points[action_point_id]
        except KeyError:
            raise CachedProjectException("Action point not found")

    def logic_item(self, logic_item_id: str) -> LogicItem:

        try:
            return self._logic_items[logic_item_id]
        except KeyError:
            raise CachedProjectException("LogicItem not found.")

    def constant(self, constant_id: str) -> ProjectConstant:

        try:
            return self._constants[constant_id]
        except KeyError:
            raise CachedProjectException("Constant not found.")


class UpdateableCachedProject(CachedProject):

    def update_modified(self) -> None:
        self._int_modified = datetime.now(tz=timezone.utc)

    @property
    def has_changes(self) -> bool:

        if self._int_modified is None:
            return False

        if self.modified is None:
            return True

        return self._int_modified > self.modified

    def upsert_action(self, ap_id: str, action: Action) -> None:

        ap = self.action_point(ap_id)

        if action.id in self._actions.data:
            assert self._actions.parent[action.id] == ap
            self._actions.data[action.id] = action
        else:
            self._actions.data[action.id] = action
            self._actions.parent[action.id] = ap
        self.update_modified()

    def remove_action(self, action_id: str) -> Action:

        try:
            action = self._actions.data.pop(action_id)
            del self._actions.parent[action_id]  # TODO KeyError here might be probably ignored? (it should not happen)
        except KeyError as e:
            raise CachedProjectException("Action not found.") from e
        self.update_modified()
        return action

    def upsert_orientation(self, ap_id: str, orientation: NamedOrientation) -> None:

        ap = self.action_point(ap_id)

        if orientation.id in self._orientations.data:
            assert self._orientations.parent[orientation.id] == ap
            self._orientations.data[orientation.id] = orientation
        else:
            self._orientations.data[orientation.id] = orientation
            self._orientations.parent[orientation.id] = ap
        self.update_modified()

    def remove_orientation(self, orientation_id: str) -> NamedOrientation:

        try:
            ori = self._orientations.data.pop(orientation_id)
            del self._orientations.parent[orientation_id]
        except KeyError as e:
            raise CachedProjectException("Orientation not found.") from e
        self.update_modified()
        return ori

    def upsert_joints(self, ap_id: str, joints: ProjectRobotJoints) -> None:

        ap = self.action_point(ap_id)

        if joints.id in self._joints.data:
            assert self._joints.parent[joints.id] == ap
            self._joints.data[joints.id] = joints
        else:
            self._joints.data[joints.id] = joints
            self._joints.parent[joints.id] = ap
        self.update_modified()

    def remove_joints(self, joints_id: str) -> ProjectRobotJoints:

        try:
            joints = self._joints.data.pop(joints_id)
            del self._joints.parent[joints_id]
        except KeyError as e:
            raise CachedProjectException("Joints not found.") from e
        self.update_modified()
        return joints

    def upsert_action_point(self, ap_id: str, name: str, position: Position, parent: Optional[str] = None) \
            -> ProjectActionPoint:

        try:
            ap = self.action_point(ap_id)
            ap.name = name
            ap.position = position
            ap.parent = parent
        except CachedProjectException:
            ap = ProjectActionPoint(ap_id, name, position, parent)
            self._action_points[ap_id] = ap
        self.update_modified()
        return ap

    def remove_action_point(self, ap_id: str) -> ProjectActionPoint:

        ap = self.action_point(ap_id)

        for action in ap.actions:
            self.remove_action(action.id)
        for joints in ap.robot_joints:
            self.remove_joints(joints.id)
        for ori in ap.orientations:
            self.remove_orientation(ori.id)

        del self._action_points[ap_id]
        self.update_modified()
        return ap

    def upsert_logic_item(self, logic_item: LogicItem) -> None:
        self._logic_items[logic_item.id] = logic_item
        self.update_modified()

    def remove_logic_item(self, logic_item_id: str) -> LogicItem:

        try:
            logic_item = self._logic_items.pop(logic_item_id)
        except KeyError as e:
            raise CachedProjectException("Logic item not found.") from e
        self.update_modified()
        return logic_item

    def clear_logic(self) -> None:

        self._logic_items.clear()
        self.update_modified()

    def upsert_constant(self, const: ProjectConstant) -> None:
        self._constants[const.id] = const
        self.update_modified()

    def remove_constant(self, const_id: str) -> ProjectConstant:

        try:
            const = self._constants.pop(const_id)
        except KeyError as e:
            raise CachedProjectException("Constant not found.") from e
        self.update_modified()
        return const



from arcor2.server.rpc import execution
from arcor2.server.rpc import objects
from arcor2.server.rpc import project
from arcor2.server.rpc import robot
from arcor2.server.rpc import scene

__all__ = [
    execution.__name__,
    objects.__name__,
    project.__name__,
    robot.__name__,
    scene.__name__
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import base64
import json
import os
import shutil
import tempfile
import uuid
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from queue import Queue
from threading import Thread
from typing import Dict, List, Optional, TYPE_CHECKING, Tuple, Union

from apispec import APISpec  # type: ignore

from apispec_webframeworks.flask import FlaskPlugin  # type: ignore

from dataclasses_jsonschema import JsonSchemaMixin
from dataclasses_jsonschema.apispec import DataclassesPlugin

from flask import Flask, Response, jsonify, request, send_file

from flask_cors import CORS  # type: ignore

from flask_swagger_ui import get_swaggerui_blueprint  # type: ignore

import websocket  # type: ignore

from werkzeug.utils import secure_filename

import arcor2
from arcor2.data import common, events, execution, rpc
from arcor2.data.helpers import EVENT_MAPPING, RPC_MAPPING
from arcor2.nodes.execution import PORT as MANAGER_PORT
from arcor2.settings import PROJECT_PATH

PORT = int(os.getenv("ARCOR2_EXECUTION_PROXY_PORT", 5009))
SERVICE_NAME = "ARCOR2 Execution Service Proxy"


RespT = Union[Response, Tuple[str, int]]


class ExecutionState(Enum):

    Undefined: str = "Undefined"
    Running: str = "Running"
    Completed: str = "Completed"
    Faulted: str = "Faulted"
    Paused: str = "Paused"


@dataclass
class SummaryPackage(JsonSchemaMixin):

    id: str
    name: Optional[str] = None
    projectId: Optional[str] = None
    created: Optional[datetime] = None
    executed: Optional[datetime] = None


@dataclass
class ExecutionInfo(JsonSchemaMixin):

    state: ExecutionState
    activePackageId: Optional[str] = None
    exceptionMessage: Optional[str] = None


# Create an APISpec
spec = APISpec(
    title=SERVICE_NAME,
    version=arcor2.version(),
    openapi_version="3.0.2",
    plugins=[FlaskPlugin(), DataclassesPlugin()],
)

app = Flask(__name__)
CORS(app)

ws: Optional[websocket.WebSocket] = None

if TYPE_CHECKING:
    ReqQueue = Queue[rpc.common.Request]  # this is only processed by mypy
    RespQueue = Queue[rpc.common.Response]
else:
    ReqQueue = Queue  # this is not seen by mypy but will be executed at runtime.
    RespQueue = Queue


rpc_request_queue: ReqQueue = ReqQueue()
rpc_responses: Dict[int, RespQueue] = {}

package_state: Optional[common.PackageState] = None
package_info: Optional[execution.PackageInfo] = None
exception_message: Optional[str] = None


def ws_thread() -> None:

    global package_info
    global package_state
    global exception_message
    assert ws

    while True:

        data = json.loads(ws.recv())

        if "event" in data:

            evt = EVENT_MAPPING[data["event"]].from_dict(data)

            if isinstance(evt, events.PackageInfoEvent):
                package_info = evt.data
            elif isinstance(evt, events.PackageStateEvent):
                package_state = evt.data

                if package_state.state == common.PackageStateEnum.RUNNING:
                    exception_message = None

            elif isinstance(evt, events.ProjectExceptionEvent):
                exception_message = evt.data.message

        elif "response" in data:
            resp = RPC_MAPPING[data["response"]][1].from_dict(data)
            rpc_responses[resp.id].put(resp)


def call_rpc(req: rpc.common.Request) -> rpc.common.Response:

    assert req.id not in rpc_responses
    assert ws

    rpc_responses[req.id] = RespQueue(maxsize=1)
    ws.send(req.to_json())
    resp = rpc_responses[req.id].get()
    del rpc_responses[req.id]
    return resp


def get_id() -> int:
    return uuid.uuid4().int


def allowed_file(filename):
    return '.' in filename and \
           filename.rsplit('.', 1)[1].lower() == "zip"


@app.route("/packages/<string:packageId>", methods=['PUT'])
def put_package(packageId: str) -> RespT:  # noqa
    """Put package
        ---
        put:
            description: Upload/update execution package.
            tags:
               - Packages
            parameters:
                - in: path
                  name: packageId
                  schema:
                    type: string
                  required: true
                  description: unique ID
            requestBody:
                  content:
                    multipart/form-data:
                      schema:
                        type: object
                        required:
                            - executionPackage
                        properties:
                          # 'file' will be the field name in this multipart request
                          executionPackage:
                            type: string
                            format: binary
            responses:
                200:
                  description: Ok
                501:
                  description: Contains array of errors.
                  content:
                    application/json:
                      schema:
                        type: array
                        items:
                          type: string

    """

    file = request.files['executionPackage']
    file_name = secure_filename(file.filename)

    with tempfile.TemporaryDirectory() as tmpdirname:
        file_path = os.path.join(tmpdirname, file_name)
        file.save(file_path)

        with open(file_path, "rb") as zip_file:
            b64_bytes = base64.b64encode(zip_file.read())
            b64_str = b64_bytes.decode()

    resp = call_rpc(rpc.execution.UploadPackageRequest(
        id=get_id(),
        args=rpc.execution.UploadPackageArgs(packageId, b64_str)))

    if resp.result:
        return "ok", 200
    else:
        return jsonify(resp.messages), 501


@app.route("/packages/<string:packageId>", methods=['GET'])
def get_package(packageId: str) -> RespT:  # noqa
    """Get execution package.
                ---
                get:
                  description: Get zip file with execution package.
                  tags:
                    - Packages
                  parameters:
                    - in: path
                      name: packageId
                      schema:
                        type: string
                      required: true
                      description: unique ID
                  responses:
                    200:
                      description: Ok
                      content:
                        application/zip:
                            schema:
                              type: string
                              format: binary
                              example: The archive of execution package (.zip)
                    404:
                        description: Package ID was not found.
                """

    package_path = os.path.join(PROJECT_PATH, packageId)

    if not os.path.exists(package_path):
        return "Not found", 404

    with tempfile.TemporaryDirectory() as tmpdirname:

        archive_path = os.path.join(tmpdirname, packageId)
        shutil.make_archive(archive_path, 'zip', package_path)

        return send_file(archive_path + ".zip", as_attachment=True, cache_timeout=0)


@app.route("/packages", methods=['GET'])
def get_packages() -> RespT:
    """Gets summary for all stored execution packages.
                ---
                get:
                  description: Summary.
                  tags:
                    - Packages
                  responses:
                    200:
                      description: Ok
                      content:
                        application/json:
                          schema:
                            type: array
                            items:
                              $ref: SummaryPackage
                """

    resp = call_rpc(rpc.execution.ListPackagesRequest(id=get_id()))
    assert isinstance(resp, rpc.execution.ListPackagesResponse)

    ret: List[Dict] = []

    for pck in resp.data:
        sp = SummaryPackage(pck.id)
        sp.name = pck.package_meta.name
        sp.created = pck.package_meta.built
        sp.projectId = pck.project_id
        sp.executed = pck.package_meta.executed
        ret.append(sp.to_dict())

    return jsonify(ret), 200


@app.route("/packages/<string:packageId>", methods=['DELETE'])
def delete_package(packageId: str) -> RespT:  # noqa
    """Delete package.
            ---
            delete:
              description: Delete package.
              tags:
                - Packages
              parameters:
                - in: path
                  name: packageId
                  schema:
                    type: string
                  required: true
                  description: unique ID
              responses:
                200:
                  description: Ok
                501:
                    description: Contains array of errors.
                    content:
                      application/json:
                        schema:
                          type: array
                          items:
                            type: string
            """

    resp = call_rpc(rpc.execution.DeletePackageRequest(id=get_id(), args=rpc.common.IdArgs(id=packageId)))

    if resp.result:
        return "ok", 200
    else:
        return jsonify(resp.messages), 501


@app.route("/packages/<string:packageId>/start", methods=['PUT'])
def package_start(packageId: str) -> RespT:  # noqa
    """Run project
            ---
            put:
              description: Start execution of the execution package.
              tags:
                - Packages
              parameters:
                - in: path
                  name: packageId
                  schema:
                    type: string
                  required: true
                  description: unique ID
              responses:
                200:
                  description: Ok
                501:
                    description: Contains array of errors.
                    content:
                      application/json:
                        schema:
                          type: array
                          items:
                            type: string
            """

    resp = call_rpc(rpc.execution.RunPackageRequest(id=get_id(), args=rpc.execution.RunPackageArgs(id=packageId)))

    if resp.result:
        return "ok", 200
    else:
        return jsonify(resp.messages), 501


@app.route("/packages/stop", methods=['PUT'])
def packages_stop() -> RespT:
    """Stops running project
            ---
            put:
              description: Stops execution of the given package.
              tags:
                - Packages
              responses:
                200:
                  description: Ok
                501:
                    description: Contains array of errors.
                    content:
                      application/json:
                        schema:
                          type: array
                          items:
                            type: string
            """

    resp = call_rpc(rpc.execution.StopPackageRequest(id=get_id()))

    if resp.result:
        return "ok", 200
    else:
        return jsonify(resp.messages), 501


@app.route("/packages/pause", methods=['PUT'])
def packages_pause() -> RespT:
    """Pauses running package.
            ---
            put:
              description: Pause execution of the given package.
              tags:
                - Packages
              responses:
                200:
                  description: Ok
                501:
                    description: Contains array of errors.
                    content:
                      application/json:
                        schema:
                          type: array
                          items:
                            type: string
            """

    resp = call_rpc(rpc.execution.PausePackageRequest(id=get_id()))

    if resp.result:
        return "ok", 200
    else:
        return jsonify(resp.messages), 501


@app.route("/packages/resume", methods=['PUT'])
def packages_resume() -> RespT:
    """Resumes running package.
            ---
            put:
              description: Resumes execution of the given package.
              tags:
                - Packages
              responses:
                200:
                  description: Ok
                501:
                    description: Contains array of errors.
                    content:
                      application/json:
                        schema:
                          type: array
                          items:
                            type: string
            """

    resp = call_rpc(rpc.execution.ResumePackageRequest(id=get_id()))

    if resp.result:
        return "ok", 200
    else:
        return jsonify(resp.messages), 501


@app.route("/packages/executioninfo", methods=['GET'])
def packages_executioninfo() -> RespT:
    """/packages/executioninfo
            ---
            get:
              description: /packages/executioninfo
              tags:
                - Packages
              responses:
                200:
                  description: Ok
                  content:
                    application/json:
                      schema:
                        $ref: ExecutionInfo
                404:
                    description: No project running
            """

    if package_state is None or package_state.state == common.PackageStateEnum.UNDEFINED:
        ret = ExecutionInfo(ExecutionState.Undefined)
    elif package_state.state == common.PackageStateEnum.RUNNING:
        ret = ExecutionInfo(ExecutionState.Running, package_state.package_id)
    elif package_state.state == common.PackageStateEnum.PAUSED:
        ret = ExecutionInfo(ExecutionState.Paused, package_state.package_id)
    elif package_state.state == common.PackageStateEnum.STOPPED:

        if exception_message:
            ret = ExecutionInfo(ExecutionState.Faulted, package_state.package_id, exception_message)
        else:
            ret = ExecutionInfo(ExecutionState.Completed, package_state.package_id)
    else:
        return "Unhandled state", 501

    return jsonify(ret.to_dict()), 200


@app.route("/swagger/api/swagger.json", methods=["GET"])
def get_swagger() -> str:
    return json.dumps(spec.to_dict())


spec.components.schema(SummaryPackage.__name__, schema=SummaryPackage)
spec.components.schema(ExecutionInfo.__name__, schema=ExecutionInfo)


with app.test_request_context():
    spec.path(view=put_package)
    spec.path(view=get_package)
    spec.path(view=get_packages)
    spec.path(view=delete_package)
    spec.path(view=package_start)
    spec.path(view=packages_stop)
    spec.path(view=packages_pause)
    spec.path(view=packages_resume)
    spec.path(view=packages_executioninfo)


def main() -> None:

    parser = argparse.ArgumentParser(description=SERVICE_NAME)
    parser.add_argument('-s', '--swagger', action="store_true", default=False)
    args = parser.parse_args()

    if args.swagger:
        print(spec.to_yaml())
        return

    global ws
    ws = websocket.create_connection(f"ws://0.0.0.0:{MANAGER_PORT}", enable_multithread=True)

    thread = Thread(target=ws_thread, daemon=True)
    thread.start()

    SWAGGER_URL = "/swagger"

    swaggerui_blueprint = get_swaggerui_blueprint(
        SWAGGER_URL,  # Swagger UI static files will be mapped to '{SWAGGER_URL}/dist/'
        "./api/swagger.json"
    )

    # Register blueprint at URL
    app.register_blueprint(swaggerui_blueprint, url_prefix=SWAGGER_URL)

    app.run(host='0.0.0.0', port=PORT)


if __name__ == '__main__':
    main()



import inspect
import os
from typing import List, Set, Type

from typed_ast.ast3 import AST

import arcor2.helpers as hlp
from arcor2.data import common, events, robot
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types import utils as otu
from arcor2.object_types.abstract import Robot
from arcor2.server import globals as glob, notifications as notif, objects_services_actions as osa
from arcor2.source.utils import function_implemented


class RobotPoseException(Arcor2Exception):
    pass


# TODO how to prevent execution of robot (move) action and e.g. moveToAp?
move_in_progress: Set[str] = set()  # set of robot IDs that are moving right now


async def get_end_effectors(robot_id: str) -> Set[str]:
    """
    :param robot_id:
    :return: IDs of existing end effectors.
    """

    robot_inst = await osa.get_robot_instance(robot_id)
    return await hlp.run_in_executor(robot_inst.get_end_effectors_ids)


async def get_grippers(robot_id: str) -> Set[str]:
    """
    :param robot_id:
    :return: IDs of existing grippers.
    """

    robot_inst = await osa.get_robot_instance(robot_id)
    return await hlp.run_in_executor(robot_inst.grippers)


async def get_suctions(robot_id: str) -> Set[str]:
    """
    :param robot_id:
    :return: IDs of existing suctions.
    """

    robot_inst = await osa.get_robot_instance(robot_id)
    return await hlp.run_in_executor(robot_inst.suctions)


async def get_end_effector_pose(robot_id: str, end_effector: str) -> common.Pose:
    """
    :param robot_id:
    :param end_effector:
    :return: Global pose
    """

    robot_inst = await osa.get_robot_instance(robot_id, end_effector)
    return await hlp.run_in_executor(robot_inst.get_end_effector_pose, end_effector)


async def get_robot_joints(robot_id: str) -> List[common.Joint]:
    """
    :param robot_id:
    :return: List of joints
    """

    robot_inst = await osa.get_robot_instance(robot_id)
    return await hlp.run_in_executor(robot_inst.robot_joints)


def feature(tree: AST, robot_type: Type[Robot], func_name: str) -> bool:

    if not function_implemented(tree, func_name):
        return False

    sign = inspect.signature(getattr(robot_type, func_name))
    return inspect.signature(getattr(Robot, func_name)) == sign


async def get_robot_meta(obj_type: otu.ObjectTypeData) -> None:

    if obj_type.meta.disabled:
        raise Arcor2Exception("Disabled object type.")

    obj_type.robot_meta = robot.RobotMeta(obj_type.meta.type)

    tree = obj_type.ast
    assert tree is not None
    assert obj_type.type_def is not None
    assert issubclass(obj_type.type_def, Robot)

    obj_type.robot_meta.features.move_to_pose = feature(tree, obj_type.type_def, Robot.move_to_pose.__name__)
    obj_type.robot_meta.features.move_to_joints = feature(tree, obj_type.type_def, Robot.move_to_joints.__name__)
    obj_type.robot_meta.features.stop = feature(tree, obj_type.type_def, Robot.stop.__name__)

    if issubclass(obj_type.type_def, Robot) and obj_type.type_def.urdf_package_path:
        obj_type.robot_meta.urdf_package_filename = os.path.split(obj_type.type_def.urdf_package_path)[1]


async def stop(robot_id: str) -> None:

    global move_in_progress

    if robot_id not in move_in_progress:
        raise Arcor2Exception("Robot is not moving.")

    robot_inst = await osa.get_robot_instance(robot_id)

    try:
        await hlp.run_in_executor(robot_inst.stop)
    except NotImplementedError as e:
        raise Arcor2Exception from e

    move_in_progress.remove(robot_id)


async def check_robot_before_move(robot_id: str) -> None:

    if robot_id in move_in_progress:
        raise Arcor2Exception("Robot is moving.")

    if glob.RUNNING_ACTION:

        assert glob.PROJECT
        action = glob.PROJECT.action(glob.RUNNING_ACTION)
        obj_id_str, _ = action.parse_type()

        if robot_id == obj_id_str:
            raise Arcor2Exception("Robot is executing action.")


async def _move_to_pose(robot_id: str, end_effector_id: str, pose: common.Pose, speed: float) -> None:
    # TODO newly connected interface should be notified somehow (general solution for such cases would be great!)

    robot_inst = await osa.get_robot_instance(robot_id, end_effector_id)
    move_in_progress.add(robot_id)

    try:
        await hlp.run_in_executor(robot_inst.move_to_pose, end_effector_id, pose, speed)
    except (NotImplementedError, Arcor2Exception) as e:
        glob.logger.error(f"Robot movement failed with: {str(e)}")
        move_in_progress.remove(robot_id)
        raise Arcor2Exception from e

    move_in_progress.remove(robot_id)


async def move_to_pose(robot_id: str, end_effector_id: str, pose: common.Pose, speed: float) -> None:

    await notif.broadcast_event(events.RobotMoveToPoseEvent(
        data=events.RobotMoveToPoseData(events.MoveEventType.START, robot_id, end_effector_id, pose)))

    try:
        await _move_to_pose(robot_id, end_effector_id, pose, speed)

    except Arcor2Exception as e:
        await notif.broadcast_event(events.RobotMoveToPoseEvent(
            data=events.RobotMoveToPoseData(events.MoveEventType.FAILED, robot_id, end_effector_id, pose,
                                            message=str(e))))
        return

    await notif.broadcast_event(events.RobotMoveToPoseEvent(
        data=events.RobotMoveToPoseData(events.MoveEventType.END, robot_id, end_effector_id, pose)))


async def move_to_ap_orientation(robot_id: str, end_effector_id: str, pose: common.Pose, speed: float,
                                 orientation_id: str) -> None:

    await notif.broadcast_event(events.RobotMoveToActionPointOrientationEvent(
        data=events.RobotMoveToActionPointOrientationData(
            events.MoveEventType.START, robot_id, end_effector_id, orientation_id)))

    try:
        await _move_to_pose(robot_id, end_effector_id, pose, speed)

    except Arcor2Exception as e:
        await notif.broadcast_event(events.RobotMoveToActionPointOrientationEvent(
            data=events.RobotMoveToActionPointOrientationData(
                events.MoveEventType.FAILED, robot_id, end_effector_id, orientation_id, message=str(e))))
        return

    await notif.broadcast_event(events.RobotMoveToActionPointOrientationEvent(
        data=events.RobotMoveToActionPointOrientationData(
            events.MoveEventType.END, robot_id, end_effector_id, orientation_id)))


async def _move_to_joints(robot_id: str, joints: List[common.Joint], speed: float) -> None:

    # TODO newly connected interface should be notified somehow (general solution for such cases would be great!)

    robot_inst = await osa.get_robot_instance(robot_id)

    move_in_progress.add(robot_id)

    try:
        await hlp.run_in_executor(robot_inst.move_to_joints, joints, speed)
    except (NotImplementedError, Arcor2Exception) as e:
        glob.logger.error(f"Robot movement failed with: {str(e)}")

        move_in_progress.remove(robot_id)
        raise Arcor2Exception from e

    move_in_progress.remove(robot_id)


async def move_to_joints(robot_id: str, joints: List[common.Joint], speed: float) -> None:

    await notif.broadcast_event(events.RobotMoveToJointsEvent(
        data=events.RobotMoveToJointsData(events.MoveEventType.START, robot_id, joints)))

    try:

        await _move_to_joints(robot_id, joints, speed)

    except Arcor2Exception as e:

        await notif.broadcast_event(events.RobotMoveToJointsEvent(
            data=events.RobotMoveToJointsData(events.MoveEventType.FAILED, robot_id, joints, message=str(e))))

        return

    await notif.broadcast_event(events.RobotMoveToJointsEvent(
        data=events.RobotMoveToJointsData(events.MoveEventType.END, robot_id, joints)))


async def move_to_ap_joints(robot_id: str, joints: List[common.Joint], speed: float, joints_id: str) -> None:

    await notif.broadcast_event(events.RobotMoveToActionPointJointsEvent(
        data=events.RobotMoveToActionPointJointsData(events.MoveEventType.START, robot_id, joints_id)))

    try:

        await _move_to_joints(robot_id, joints, speed)

    except Arcor2Exception as e:

        await notif.broadcast_event(events.RobotMoveToActionPointJointsEvent(
            data=events.RobotMoveToActionPointJointsData(events.MoveEventType.FAILED, robot_id, joints_id,
                                                         message=str(e))))

        return

    await notif.broadcast_event(events.RobotMoveToActionPointJointsEvent(
        data=events.RobotMoveToActionPointJointsData(events.MoveEventType.END, robot_id, joints_id)))



import inspect
import os
from typing import Type

import horast

from typed_ast.ast3 import AST, AnnAssign, Assign, ClassDef, ImportFrom, Index, Load, Module, Name, NameConstant,\
    Pass, Store, Str, Subscript, alias

import arcor2.object_types
from arcor2.data.object_type import ObjectTypeMeta
from arcor2.helpers import camel_case_to_snake_case
from arcor2.object_types.abstract import Generic
from arcor2.object_types.utils import built_in_types_names, meta_from_def, object_actions
from arcor2.parameter_plugins import TYPE_TO_PLUGIN
from arcor2.source.utils import find_function, get_name, get_name_attr


def prepare_object_types_dir(path: str, module: str) -> None:

    full_path = os.path.join(path, module)

    if not os.path.exists(full_path):
        os.makedirs(full_path)

    open(os.path.join(full_path, "__init__.py"), "w").close()


def check_object_type(type_def: Type[Generic]) -> None:
    """
    Checks whether the object type source is a valid one.
    :param object_type_source:
    :return:
    """

    meta_from_def(type_def, False)
    object_actions(TYPE_TO_PLUGIN, type_def, horast.parse(inspect.getsource(type_def)))


def fix_object_name(object_id: str) -> str:

    return camel_case_to_snake_case(object_id).replace(' ', '_')


# TODO could this be done like this https://stackoverflow.com/a/9269964/3142796 ??
def new_object_type(parent: ObjectTypeMeta, child: ObjectTypeMeta) -> AST:

    assert parent.type == child.base

    tree = Module(body=[], type_ignores=[])

    if parent.type in built_in_types_names():
        import_from = arcor2.object_types.abstract.__name__
    else:
        import_from = f".{camel_case_to_snake_case(parent.type)}"

    tree.body.append(ImportFrom(module=import_from,
                                names=[alias(name=parent.type, asname=None)],
                                level=0))

    c = ClassDef(
        name=child.type,
        bases=[get_name(parent.type)],
        keywords=[],
        body=[
            Assign(
                targets=[Name(id='_ABSTRACT', ctx=Store())],
                value=NameConstant(value=False),
                type_comment=None)
        ],
        decorator_list=[])

    # TODO add docstring with description (if provided)
    c.body.append(Pass())

    tree.body.append(c)

    return tree


def object_instance_from_res(tree: Module, object_name: str, object_id: str, cls_name: str) -> None:

    main_body = find_function("main", tree).body
    last_assign_idx = -1

    for body_idx, body_item in enumerate(main_body):

        if isinstance(body_item, (Assign, AnnAssign)):
            last_assign_idx = body_idx

    assign = AnnAssign(
        target=Name(
            id=fix_object_name(object_name),
            ctx=Store()),
        annotation=Name(
            id=cls_name,
            ctx=Load()),
        value=Subscript(
            value=get_name_attr('res', "objects"),
            slice=Index(value=Str(
                s=object_id,
                kind='')),
            ctx=Load()),
        simple=1)

    main_body.insert(last_assign_idx + 1, assign)



import copy
import json
from typing import Any, List

from arcor2 import transformations as tr
from arcor2.cached import CachedProject as CProject, CachedScene as CScene
from arcor2.data.common import Pose
from arcor2.parameter_plugins.base import ParameterPlugin, TypesDict
from arcor2.parameter_plugins.list import ListParameterPlugin, get_type_name


class PosePlugin(ParameterPlugin):

    @classmethod
    def type(cls) -> Any:
        return Pose

    @classmethod
    def value(cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str) \
            -> Pose:

        action = project.action(action_id)
        param = action.parameter(parameter_id)
        ori_id: str = cls.param_value(param)

        ap, ori = project.ap_and_orientation(ori_id)
        return Pose(ap.position, ori.orientation)

    @classmethod
    def execution_value(cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str,
                        parameter_id: str) -> Pose:

        action = project.action(action_id)
        param = action.parameter(parameter_id)
        ori_id: str = cls.param_value(param)

        ap, _ = project.ap_and_orientation(ori_id)

        copy_of_ap = copy.deepcopy(ap)
        tr.make_relative_ap_global(scene, project, copy_of_ap)

        return Pose(copy_of_ap.position, project.orientation(ori_id).orientation)

    @classmethod
    def value_to_json(cls, value: Pose) -> str:
        return value.to_json()

    @classmethod
    def uses_orientation(cls, project: CProject, action_id: str, parameter_id: str, orientation_id: str) -> bool:

        action = project.action(action_id)
        param = action.parameter(parameter_id)
        return orientation_id == cls.param_value(param)


class PoseListPlugin(ListParameterPlugin):

    @classmethod
    def type(cls):
        return List[Pose]

    @classmethod
    def type_name(cls) -> str:
        return get_type_name(PosePlugin)

    @classmethod
    def value(cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str) \
            -> List[Pose]:

        ret: List[Pose] = []

        ap, action = project.action_point_and_action(action_id)
        parameter = action.parameter(parameter_id)

        for orientation_id in cls.param_value_list(parameter):
            ret.append(Pose(ap.position, project.orientation(orientation_id).orientation))

        return ret

    @classmethod
    def execution_value(cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str,
                        parameter_id: str) -> List[Pose]:

        ap, action = project.action_point_and_action(action_id)

        if not ap.parent:
            return cls.value(type_defs, scene, project, action_id, parameter_id)

        parameter = action.parameter(parameter_id)
        ret: List[Pose] = []

        copy_of_ap = copy.deepcopy(ap)
        tr.make_relative_ap_global(scene, project, copy_of_ap)

        for orientation_id in cls.param_value_list(parameter):
            ret.append(Pose(ap.position, project.orientation(orientation_id).orientation))

        return ret

    @classmethod
    def value_to_json(cls, value: List[Pose]) -> str:
        return json.dumps([v.to_json() for v in value])

    @classmethod
    def uses_orientation(cls, project: CProject, action_id: str, parameter_id: str, orientation_id: str) -> bool:

        action = project.action(action_id)
        param = action.parameter(parameter_id)

        for ori_id in cls.param_value_list(param):
            if ori_id == orientation_id:
                return True
        return False



import select
import sys
from functools import wraps
from typing import Any, Callable, Type, TypeVar, Union, cast

from arcor2.data.common import ActionState, ActionStateEnum, PackageState, PackageStateEnum
from arcor2.data.events import ActionStateEvent, Event, PackageStateEvent
from arcor2.object_types.abstract import Generic
from arcor2.object_types.utils import iterate_over_actions

HANDLE_ACTIONS = True


def patch_object_actions(type_def: Type[Generic]) -> None:
    """
    Dynamically adds @action decorator to the methods with assigned ActionMetadata.
    :param type_def:
    :return:
    """

    for method_name, method in iterate_over_actions(type_def):
        setattr(type_def, method_name, action(method))


try:
    # Windows solution
    import msvcrt
    import time

    def read_stdin(timeout: float = 0.0) -> Union[str, None]:

        time_to_end = time.monotonic() + timeout

        while True:

            x = msvcrt.kbhit()  # type: ignore
            if x:
                return msvcrt.getch().decode()  # type: ignore

            if not timeout or time.monotonic() > time_to_end:
                return None

            time.sleep(timeout / 100.0)


except ImportError:

    # Linux solution
    def read_stdin(timeout: float = 0.0) -> Union[str, None]:

        if select.select([sys.stdin], [], [], timeout)[0]:
            return sys.stdin.readline().strip()
        return None


def handle_action(inst: Generic, f: Callable[..., Any], where: ActionStateEnum) -> None:

    print_event(ActionStateEvent(data=ActionState(inst.id, f.__name__, where)))

    ctrl_cmd = read_stdin()

    if ctrl_cmd == "p":
        print_event(PackageStateEvent(data=PackageState(PackageStateEnum.PAUSED)))
        while True:
            ctrl_cmd = read_stdin(0.1)
            if ctrl_cmd == "r":
                print_event(PackageStateEvent(data=PackageState(PackageStateEnum.RUNNING)))
                break


def print_event(event: Event) -> None:
    """
    Used from main script to print event as JSON.
    """

    print(event.to_json())
    sys.stdout.flush()


F = TypeVar('F', bound=Callable[..., Any])


def action(f: F) -> F:

    @wraps(f)
    def wrapper(*args: Union[Generic, Any], **kwargs: Any) -> Any:

        # automagical overload for dictionary (allow to get rid of ** in script).
        if len(args) == 2 and isinstance(args[1], dict) and not kwargs:
            kwargs = args[1]
            args = (args[0],)

        if not action.inside_composite and HANDLE_ACTIONS:  # type: ignore
            handle_action(args[0], f, ActionStateEnum.BEFORE)

        if wrapper.__action__.composite:  # type: ignore # TODO and not step_into
            action.inside_composite = f  # type: ignore

        res = f(*args, **kwargs)

        if action.inside_composite == f:  # type: ignore
            action.inside_composite = None  # type: ignore

        if not action.inside_composite and HANDLE_ACTIONS:  # type: ignore
            handle_action(args[0], f, ActionStateEnum.AFTER)

        return res

    return cast(F, wrapper)


action.inside_composite = None  # type: ignore



import os
from typing import Set

from arcor2 import rest
from arcor2.data.common import Pose
from arcor2.data.object_type import MeshFocusAction, Model3dType, Models
from arcor2.exceptions import Arcor2Exception

URL = os.getenv("ARCOR2_SCENE_SERVICE_URL", "http://0.0.0.0:5013")


class SceneServiceException(Arcor2Exception):
    pass


def upsert_collision(model: Models, pose: Pose) -> None:
    """
    Adds arbitrary collision model to the collision scene.
    :param model: Box, Sphere, Cylinder, Mesh
    :param pose: Pose of the collision object.
    :return:

    Example usage:

    >>> from arcor2.data.object_type import Box
    >>> from arcor2.data.common import Pose, Position, Orientation
    >>> box = Box("UniqueBoxId", 0.1, 0.1, 0.1)
    >>> scene_service.upsert_collision(box, Pose(Position(1, 0, 0), Orientation(0, 0, 0, 1)))

    """

    model_id = model.id
    params = model.to_dict()
    del params["id"]
    params[model.__class__.__name__.lower() + "Id"] = model_id

    # TODO temporary hack
    if model.type() == Model3dType.MESH:
        params["mesh_scale_x"] = 1.0
        params["mesh_scale_y"] = 1.0
        params["mesh_scale_z"] = 1.0
        params["transform_id"] = "world"

    rest.put(f"{URL}/collisions/{model.type().value.lower()}", pose, params)


@rest.handle_exceptions(SceneServiceException)
def delete_collision_id(collision_id: str) -> None:
    rest.delete(f"{URL}/collisions/{collision_id}")


@rest.handle_exceptions(SceneServiceException)
def collision_ids() -> Set[str]:
    return set(rest.get_list_primitive(f"{URL}/collisions", str))


@rest.handle_exceptions(SceneServiceException)
def focus(mfa: MeshFocusAction) -> Pose:
    return rest.put(f"{URL}/utils/focus", mfa, data_cls=Pose)


def delete_all_collisions() -> None:

    for cid in collision_ids():
        delete_collision_id(cid)


__all__ = [
    SceneServiceException.__name__,
    upsert_collision.__name__,
    delete_collision_id.__name__,
    collision_ids.__name__,
    focus.__name__,
    delete_all_collisions.__name__
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import json
import os
import uuid
from datetime import datetime, timezone
from typing import Dict, Tuple, Union, cast

from apispec import APISpec  # type: ignore

from apispec_webframeworks.flask import FlaskPlugin  # type: ignore

from dataclasses_jsonschema.apispec import DataclassesPlugin

from flask import Flask, Response, jsonify, request

from flask_cors import CORS  # type: ignore

from flask_swagger_ui import get_swaggerui_blueprint  # type: ignore

import arcor2
from arcor2.data import common, object_type
from arcor2.helpers import camel_case_to_snake_case
from arcor2.rest import convert_keys

PORT = int(os.getenv("ARCOR2_PROJECT_SERVICE_MOCK_PORT", 5012))
SERVICE_NAME = "ARCOR2 Project Service Mock"


# Create an APISpec
spec = APISpec(
    title=SERVICE_NAME,
    version=arcor2.version(),
    openapi_version="3.0.2",
    plugins=[FlaskPlugin(), DataclassesPlugin()],
)

app = Flask(__name__)
CORS(app)


def get_id() -> int:
    return uuid.uuid4().int


RespT = Union[Response, Tuple[str, int]]


SCENES: Dict[str, common.Scene] = {}
PROJECTS: Dict[str, common.Project] = {}
OBJECT_TYPES: Dict[str, object_type.ObjectType] = {}

BOXES: Dict[str, object_type.Box] = {}
CYLINDERS: Dict[str, object_type.Cylinder] = {}
SPHERES: Dict[str, object_type.Sphere] = {}


@app.route("/project", methods=['PUT'])
def put_project() -> RespT:
    """Add or update project.
        ---
        put:
            tags:
                - Project
            description: Add or update project.
            requestBody:
                  content:
                    application/json:
                      schema:
                        $ref: Project
            responses:
                200:
                  description: Ok
    """

    project = common.Project.from_dict(convert_keys(request.json, camel_case_to_snake_case))
    project.modified = datetime.now(tz=timezone.utc)
    PROJECTS[project.id] = project
    return "ok", 200


@app.route("/project/<string:id>", methods=['GET'])
def get_project(id: str) -> RespT:
    """Add or update project.
        ---
        get:
            tags:
                - Project
            summary: Gets projet by project id.
            parameters:
                - name: id
                  in: path
                  description: unique ID
                  required: true
                  schema:
                    type: string
            responses:
                200:
                  description: Ok
                  content:
                    application/json:
                        schema:
                            $ref: Project
    """

    try:
        return cast(Response, jsonify(PROJECTS[id].to_dict()))
    except KeyError:
        return "Not found", 404


@app.route("/project/<string:id>", methods=['DELETE'])
def delete_project(id: str) -> RespT:
    """Deletes project.
        ---
        delete:
            tags:
                - Project
            summary: Deletes project.
            parameters:
                - name: id
                  in: path
                  description: unique ID
                  required: true
                  schema:
                    type: string
            responses:
                200:
                  description: Ok
    """

    try:
        del PROJECTS[id]
    except KeyError:
        return "Not found", 404

    return "ok", 200


@app.route("/projects", methods=['GET'])
def get_projects() -> RespT:
    """Add or update project.
        ---
        get:
            tags:
            - Project
            summary: Gets all projects id and description.
            responses:
                '200':
                  description: Success
                  content:
                    application/json:
                      schema:
                        $ref: IdDescList
    """

    ret = common.IdDescList()

    for proj in PROJECTS.values():
        ret.items.append(common.IdDesc(proj.id, proj.name, proj.desc))

    return cast(Response, jsonify(ret.to_dict()))


@app.route("/scene", methods=['PUT'])
def put_scene() -> RespT:
    """Add or update scene.
        ---
        put:
            tags:
                - Scene
            description: Add or update scene.
            requestBody:
                  content:
                    application/json:
                      schema:
                        $ref: Scene
            responses:
                200:
                  description: Ok
    """

    scene = common.Scene.from_dict(convert_keys(request.json, camel_case_to_snake_case))
    scene.modified = datetime.now(tz=timezone.utc)
    SCENES[scene.id] = scene
    return "ok", 200


@app.route("/scene/<string:id>", methods=['GET'])
def get_scene(id: str) -> RespT:
    """Add or update scene.
        ---
        get:
            tags:
                - Scene
            summary: Gets scene by project id.
            parameters:
                - name: id
                  in: path
                  description: unique ID
                  required: true
                  schema:
                    type: string
            responses:
                200:
                  description: Ok
                  content:
                    application/json:
                        schema:
                            $ref: Scene
    """

    try:
        return cast(Response, jsonify(SCENES[id].to_dict()))
    except KeyError:
        return "Not found", 404


@app.route("/scene/<string:id>", methods=['DELETE'])
def delete_scene(id: str) -> RespT:
    """Deletes scene.
        ---
        delete:
            tags:
                - Scene
            summary: Deletes scene.
            parameters:
                - name: id
                  in: path
                  description: unique ID
                  required: true
                  schema:
                    type: string
            responses:
                200:
                  description: Ok
    """

    try:
        del SCENES[id]
    except KeyError:
        return "Not found", 404

    return "ok", 200


@app.route("/scenes", methods=['GET'])
def get_scenes() -> RespT:
    """Add or update scene.
        ---
        get:
            tags:
            - Scene
            summary: Gets all scenes id and description.
            responses:
                '200':
                  description: Success
                  content:
                    application/json:
                      schema:
                        $ref: IdDescList
    """

    ret = common.IdDescList()

    for scene in SCENES.values():
        ret.items.append(common.IdDesc(scene.id, scene.name, scene.desc))

    return cast(Response, jsonify(ret.to_dict()))


@app.route("/object_type", methods=['PUT'])
def put_object_type() -> RespT:
    """Add or update object type.
        ---
        put:
            tags:
                - ObjectType
            description: Add or update object type.
            requestBody:
                  content:
                    application/json:
                      schema:
                        $ref: ObjectType
            responses:
                200:
                  description: Ok
    """

    obj_type = object_type.ObjectType.from_dict(convert_keys(request.json, camel_case_to_snake_case))
    OBJECT_TYPES[obj_type.id] = obj_type
    return "ok", 200


@app.route("/object_types/<string:id>", methods=['GET'])
def get_object_type(id: str) -> RespT:
    """Add or update object_type.
        ---
        get:
            tags:
                - ObjectType
            summary: Gets object_type by project id.
            parameters:
                - name: id
                  in: path
                  description: unique ID
                  required: true
                  schema:
                    type: string
            responses:
                200:
                  description: Ok
                  content:
                    application/json:
                        schema:
                            $ref: ObjectType
    """

    try:
        return cast(Response, jsonify(OBJECT_TYPES[id].to_dict()))
    except KeyError:
        return "Not found", 404


@app.route("/object_type/<string:id>", methods=['DELETE'])
def delete_object_type(id: str) -> RespT:
    """Deletes object type.
        ---
        delete:
            tags:
                - ObjectType
            summary: Deletes object type.
            parameters:
                - name: id
                  in: path
                  description: unique ID
                  required: true
                  schema:
                    type: string
            responses:
                200:
                  description: Ok
    """

    try:
        del OBJECT_TYPES[id]
    except KeyError:
        return "Not found", 404

    return "ok", 200


@app.route("/object_types", methods=['GET'])
def get_object_types() -> RespT:
    """Add or update ObjectType.
        ---
        get:
            tags:
            - ObjectType
            summary: Gets all object types id and description.
            responses:
                '200':
                  description: Success
                  content:
                    application/json:
                      schema:
                        $ref: IdDescList
    """

    ret = common.IdDescList()

    for obj_type in OBJECT_TYPES.values():
        ret.items.append(common.IdDesc(obj_type.id, "", obj_type.desc))

    return cast(Response, jsonify(ret.to_dict()))


@app.route("/models/box", methods=['PUT'])
def put_box() -> RespT:
    """Add or update box.
        ---
        put:
            tags:
                - Models
            description: Add or update service type.
            requestBody:
                  content:
                    application/json:
                      schema:
                        $ref: Box
            responses:
                200:
                  description: Ok
    """

    box = object_type.Box.from_dict(convert_keys(request.json, camel_case_to_snake_case))
    BOXES[box.id] = box
    return "ok", 200


@app.route("/models/<string:id>/box", methods=['GET'])
def get_box(id: str) -> RespT:
    """Add or update box.
        ---
        get:
            tags:
                - Models
            summary: Gets ServiceType by service id.
            parameters:
                - name: id
                  in: path
                  description: unique ID
                  required: true
                  schema:
                    type: string
            responses:
                200:
                  description: Ok
                  content:
                    application/json:
                        schema:
                            $ref: Box
    """

    try:
        return cast(Response, jsonify(BOXES[id].to_dict()))
    except KeyError:
        return "Not found", 404


@app.route("/models/box", methods=['PUT'])
def put_cylinder() -> RespT:
    """Add or update box.
        ---
        put:
            tags:
                - Models
            description: Add or update service type.
            requestBody:
                  content:
                    application/json:
                      schema:
                        $ref: Cylinder
            responses:
                200:
                  description: Ok
    """

    cylinder = object_type.Cylinder.from_dict(convert_keys(request.json, camel_case_to_snake_case))
    CYLINDERS[cylinder.id] = cylinder
    return "ok", 200


@app.route("/models/<string:id>/cylinder", methods=['GET'])
def get_cylinder(id: str) -> RespT:
    """Add or update box.
        ---
        get:
            tags:
                - Models
            summary: Gets ServiceType by service id.
            parameters:
                - name: id
                  in: path
                  description: unique ID
                  required: true
                  schema:
                    type: string
            responses:
                200:
                  description: Ok
                  content:
                    application/json:
                        schema:
                            $ref: Cylinder
    """

    try:
        return cast(Response, jsonify(CYLINDERS[id].to_dict()))
    except KeyError:
        return "Not found", 404


@app.route("/models/sphere", methods=['PUT'])
def put_sphere() -> RespT:
    """Add or update sphere.
        ---
        put:
            tags:
                - Models
            description: Add or update sphere.
            requestBody:
                  content:
                    application/json:
                      schema:
                        $ref: Sphere
            responses:
                200:
                  description: Ok
    """

    sphere = object_type.Sphere.from_dict(convert_keys(request.json, camel_case_to_snake_case))
    SPHERES[sphere.id] = sphere
    return "ok", 200


@app.route("/models/<string:id>/sphere", methods=['GET'])
def get_sphere(id: str) -> RespT:
    """Add or update sphere.
        ---
        get:
            tags:
                - Models
            summary: Gets sphere by id.
            parameters:
                - name: id
                  in: path
                  description: unique ID
                  required: true
                  schema:
                    type: string
            responses:
                200:
                  description: Ok
                  content:
                    application/json:
                        schema:
                            $ref: Sphere
    """

    try:
        return cast(Response, jsonify(SPHERES[id].to_dict()))
    except KeyError:
        return "Not found", 404


@app.route("/models/<string:id>", methods=['DELETE'])
def delete_model(id: str) -> RespT:
    """Deletes model.
        ---
        delete:
            tags:
                - Models
            summary: Deletes model.
            parameters:
                - name: id
                  in: path
                  description: unique ID
                  required: true
                  schema:
                    type: string
            responses:
                200:
                  description: Ok
    """

    try:
        del BOXES[id]
    except KeyError:
        try:
            del CYLINDERS[id]
        except KeyError:
            try:
                del SPHERES[id]
            except KeyError:
                return "Not found", 404

    return "ok", 200


@app.route("/swagger/api/swagger.json", methods=["GET"])
def get_swagger() -> str:
    return json.dumps(spec.to_dict())


spec.components.schema(common.Project.__name__, schema=common.Project)
spec.components.schema(common.Scene.__name__, schema=common.Scene)
spec.components.schema(common.IdDescList.__name__, schema=common.IdDescList)
spec.components.schema(object_type.ObjectType.__name__, schema=object_type.ObjectType)
spec.components.schema(object_type.Box.__name__, schema=object_type.Box)
spec.components.schema(object_type.Cylinder.__name__, schema=object_type.Cylinder)
spec.components.schema(object_type.Sphere.__name__, schema=object_type.Sphere)


with app.test_request_context():

    spec.path(view=put_project)
    spec.path(view=get_project)
    spec.path(view=delete_project)
    spec.path(view=get_projects)

    spec.path(view=put_scene)
    spec.path(view=get_scene)
    spec.path(view=delete_scene)
    spec.path(view=get_scenes)

    spec.path(view=put_object_type)
    spec.path(view=get_object_type)
    spec.path(view=delete_object_type)
    spec.path(view=get_object_types)

    spec.path(view=put_box)
    spec.path(view=get_box)
    spec.path(view=put_cylinder)
    spec.path(view=get_cylinder)
    spec.path(view=put_sphere)
    spec.path(view=get_sphere)
    spec.path(view=delete_model)


def main() -> None:

    parser = argparse.ArgumentParser(description=SERVICE_NAME)
    parser.add_argument('-s', '--swagger', action="store_true", default=False)
    args = parser.parse_args()

    if args.swagger:
        print(spec.to_yaml())
        return

    SWAGGER_URL = "/swagger"

    swaggerui_blueprint = get_swaggerui_blueprint(
        SWAGGER_URL,  # Swagger UI static files will be mapped to '{SWAGGER_URL}/dist/'
        "./api/swagger.json"
    )

    # Register blueprint at URL
    app.register_blueprint(swaggerui_blueprint, url_prefix=SWAGGER_URL)

    app.run(host='0.0.0.0', port=PORT)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import os
from collections import defaultdict
from typing import Any, DefaultDict, Dict, Optional, Set

from aiologger import Logger  # type: ignore
from aiologger.levels import LogLevel  # type: ignore

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import helpers as hlp
from arcor2.cached import UpdateableCachedProject, UpdateableCachedScene
from arcor2.data import events
from arcor2.data.common import ActionState, CurrentAction, PackageState
from arcor2.data.execution import PackageInfo
from arcor2.nodes.build import PORT as BUILD_PORT
from arcor2.nodes.execution import PORT as EXE_PORT
from arcor2.object_types.abstract import Generic
from arcor2.object_types.utils import ObjectTypeDict


logger = Logger.with_default_handlers(name='server', formatter=hlp.aiologger_formatter(), level=LogLevel.DEBUG)
VERBOSE: bool = False

MANAGER_URL = os.getenv("ARCOR2_EXECUTION_URL", f"ws://0.0.0.0:{EXE_PORT}")
BUILDER_URL = os.getenv("ARCOR2_BUILDER_URL", f"http://0.0.0.0:{BUILD_PORT}")

PORT: int = int(os.getenv("ARCOR2_SERVER_PORT", 6789))

SCENE: Optional[UpdateableCachedScene] = None
PROJECT: Optional[UpdateableCachedProject] = None

MAIN_SCREEN: Optional[events.ShowMainScreenData] = \
    events.ShowMainScreenData(events.ShowMainScreenData.WhatEnum.ScenesList)

INTERFACES: Set[WsClient] = set()

OBJECT_TYPES: ObjectTypeDict = {}

SCENE_OBJECT_INSTANCES: Dict[str, Generic] = {}

RUNNING_ACTION: Optional[str] = None  # ID of an action that is being executed during project editing
RUNNING_ACTION_PARAMS: Optional[Dict[str, Any]] = None

PACKAGE_STATE = PackageState()
PACKAGE_INFO: Optional[PackageInfo] = None
ACTION_STATE: Optional[ActionState] = None
CURRENT_ACTION: Optional[CurrentAction] = None
TEMPORARY_PACKAGE: bool = False

RegisteredUiDict = DefaultDict[str, Set[WsClient]]

ROBOT_JOINTS_REGISTERED_UIS: RegisteredUiDict = defaultdict(lambda: set())  # robot, UIs
ROBOT_EEF_REGISTERED_UIS: RegisteredUiDict = defaultdict(lambda: set())  # robot, UIs



from typing import Set

from typed_ast.ast3 import Module, Pass

import arcor2.object_types
from arcor2.cached import CachedProject as CProject, CachedScene as CScene
from arcor2.exceptions import Arcor2Exception
from arcor2.helpers import camel_case_to_snake_case
from arcor2.logic import ActionCacheTuple
from arcor2.source import SCRIPT_HEADER, SourceException
from arcor2.source.object_types import fix_object_name
from arcor2.source.object_types import object_instance_from_res
from arcor2.source.utils import add_import, append_method_call, clean, empty_script_tree, get_name_attr,\
    main_loop_body, tree_to_str


def program_src(project: CProject, scene: CScene, built_in_objects: Set[str], add_logic: bool = True) -> str:

    tree = empty_script_tree(add_main_loop=add_logic)

    # get object instances from resources object
    for obj in scene.objects:

        if obj.type in built_in_objects:
            add_import(tree, arcor2.object_types.__name__, obj.type, try_to_import=False)
        else:
            add_import(tree, "object_types." + camel_case_to_snake_case(obj.type), obj.type, try_to_import=False)

        object_instance_from_res(tree, obj.name, obj.id, obj.type)

    if add_logic:
        add_logic_to_loop(tree, scene, project)

    return SCRIPT_HEADER + tree_to_str(tree)


def add_logic_to_loop(tree: Module, scene: CScene, project: CProject) -> None:

    loop = main_loop_body(tree)

    try:
        cache = ActionCacheTuple.from_logic_container(project)
    except Arcor2Exception as e:
        raise SourceException(e) from e

    if not cache.first_logic_item:
        raise SourceException("Can't process unfinished logic.")

    next_logic_item = cache.first_logic_item

    while True:

        if next_logic_item.end != next_logic_item.END:
            act = cache.actions[next_logic_item.end]
        else:
            break

        for logic_item in cache.logic.values():
            if logic_item.start == act.id:
                next_logic_item = logic_item
                break
        else:
            raise Arcor2Exception("Next logic item not found.")

        if len(loop) == 1 and isinstance(loop[0], Pass):
            # pass is not necessary now
            loop.clear()

        ac_obj, ac_type = act.type.split('/')

        # for scene objects, convert ID to name
        try:
            ac_obj = scene.object(ac_obj).name
        except Arcor2Exception:
            pass

        append_method_call(loop, fix_object_name(ac_obj), ac_type, [get_name_attr("res", clean(act.name))], [])



# -*- coding: utf-8 -*-

from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional, Set

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data.common import IdDesc, Pose, SceneObjectSetting
from arcor2.data.rpc.common import IdArgs, Request, Response, wo_suffix


@dataclass
class RenameArgs(JsonSchemaMixin):

    id: str
    new_name: str

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class SceneObjectUsageRequest(Request):

    args: IdArgs = field(metadata=dict(description="ID could be object or service."))
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class SceneObjectUsageResponse(Response):

    data: Set[str] = field(default_factory=set)
    response: str = field(default=SceneObjectUsageRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------

@dataclass
class AddObjectToSceneRequestArgs(JsonSchemaMixin):

    name: str
    type: str
    pose: Optional[Pose] = None
    settings: List[SceneObjectSetting] = field(default_factory=list)


@dataclass
class AddObjectToSceneRequest(Request):

    args: AddObjectToSceneRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class AddObjectToSceneResponse(Response):

    response: str = field(default=AddObjectToSceneRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RemoveFromSceneArgs(IdArgs):

    force: bool = False


@dataclass
class RemoveFromSceneRequest(Request):

    args: RemoveFromSceneArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RemoveFromSceneResponse(Response):
    response: str = field(default=RemoveFromSceneRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class SaveSceneRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class SaveSceneResponse(Response):

    response: str = field(default=SaveSceneRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class OpenSceneRequest(Request):

    args: IdArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class OpenSceneResponse(Response):

    response: str = field(default=OpenSceneRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class ListScenesRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ListScenesResponseData(IdDesc):

    modified: Optional[datetime] = None


@dataclass
class ListScenesResponse(Response):

    data: List[ListScenesResponseData] = field(default_factory=list)
    response: str = field(default=ListScenesRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class NewSceneRequestArgs(JsonSchemaMixin):

    name: str
    desc: str = field(default_factory=str)


@dataclass
class NewSceneRequest(Request):

    args: NewSceneRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class NewSceneResponse(Response):

    response: str = field(default=NewSceneRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class CloseSceneRequestArgs(JsonSchemaMixin):

    force: bool = False


@dataclass
class CloseSceneRequest(Request):

    args: CloseSceneRequestArgs = field(default_factory=CloseSceneRequestArgs)
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class CloseSceneResponse(Response):

    response: str = field(default=CloseSceneRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class UpdateObjectPoseRequestArgs(JsonSchemaMixin):

    object_id: str
    pose: Pose


@dataclass
class UpdateObjectPoseRequest(Request):

    args: UpdateObjectPoseRequestArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UpdateObjectPoseResponse(Response):

    response: str = field(default=UpdateObjectPoseRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RenameObjectRequest(Request):

    args: RenameArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RenameObjectResponse(Response):

    response: str = field(default=RenameObjectRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RenameSceneRequest(Request):

    args: RenameArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RenameSceneResponse(Response):

    response: str = field(default=RenameSceneRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class DeleteSceneRequest(Request):

    args: IdArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class DeleteSceneResponse(Response):

    data: Set[str] = field(default_factory=set)
    response: str = field(default=DeleteSceneRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class ProjectsWithSceneRequest(Request):

    args: IdArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ProjectsWithSceneResponse(Response):

    data: Set[str] = field(default_factory=set)
    response: str = field(default=ProjectsWithSceneRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class UpdateSceneDescriptionArgs(JsonSchemaMixin):

    scene_id: str
    new_description: str


@dataclass
class UpdateSceneDescriptionRequest(Request):

    args: UpdateSceneDescriptionArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UpdateSceneDescriptionResponse(Response):

    response: str = field(default=UpdateSceneDescriptionRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class CopySceneArgs(JsonSchemaMixin):

    source_id: str
    target_name: str


@dataclass
class CopySceneRequest(Request):

    args: CopySceneArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class CopySceneResponse(Response):

    data: str = ""
    response: str = field(default=CopySceneRequest.request, init=False)



import asyncio
import importlib
import json
import keyword
import logging
import os
import re
import sys
import time
import traceback
from collections import deque
from datetime import datetime, timezone
from types import ModuleType
from typing import Any, Awaitable, Callable, Coroutine, Dict, Optional, Set, Tuple, Type, TypeVar

from aiologger.formatters.base import Formatter  # type: ignore
from aiologger.levels import LogLevel  # type: ignore

from dataclasses_jsonschema import ValidationError

import semver  # type: ignore

import websockets

from arcor2.data.events import Event, ProjectExceptionEvent, ProjectExceptionEventData
from arcor2.data.execution import PackageMeta
from arcor2.data.helpers import EVENT_MAPPING, RPC_MAPPING
from arcor2.data.rpc.common import Request, Response
from arcor2.exceptions import Arcor2Exception
from arcor2.settings import PROJECT_PATH


ReqT = TypeVar("ReqT", bound=Request)
RespT = TypeVar("RespT", bound=Response)

RPC_CB = Callable[[ReqT, websockets.WebSocketServerProtocol], Coroutine[Any, Any, Optional[RespT]]]
RPC_DICT_TYPE = Dict[Type[ReqT], RPC_CB]

EventT = TypeVar("EventT", bound=Event)
EVENT_DICT_TYPE = Dict[Type[EventT], Callable[[EventT, websockets.WebSocketServerProtocol], Coroutine[Any, Any, None]]]

LOG_FORMAT = '%(name)s - %(levelname)-8s: %(message)s'


class ImportClsException(Arcor2Exception):
    pass


class TypeDefException(Arcor2Exception):
    pass


def aiologger_formatter() -> Formatter:

    return Formatter(LOG_FORMAT)


def logger_formatter() -> logging.Formatter:

    return logging.Formatter(LOG_FORMAT)


def import_cls(module_cls: str) -> Tuple[ModuleType, Type[Any]]:
    """
    Gets module and class based on string like 'module/Cls'.
    :param module_cls:
    :return:
    """

    try:
        module_name, cls_name = module_cls.split('/')
    except (IndexError, ValueError):
        raise ImportClsException("Invalid format.")

    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError:
        raise ImportClsException(f"Module '{module_name}' not found.")

    try:
        cls = getattr(module, cls_name)
    except AttributeError:
        raise ImportClsException(f"Class {cls_name} not found in module '{module_name}'.")

    return module, cls


def is_valid_identifier(value: str) -> bool:
    """
    Identifier (e.g. object id) will be used as variable name in the script - it should be in snake_case,
    not containing any special characters etc.
    :param value:
    :return:
    """

    return value.isidentifier() and not keyword.iskeyword(value) and value == camel_case_to_snake_case(value)


def is_valid_type(value: str) -> bool:
    """
    Value will be used as object type name - it should be in CamelCase,
    not containing any special characters etc.
    :param value:
    :return:
    """

    return value.isidentifier() and not keyword.iskeyword(value) and value == snake_case_to_camel_case(value)


_first_cap_re = re.compile('(.)([A-Z][a-z]+)')
_all_cap_re = re.compile('([a-z0-9])([A-Z])')


def camel_case_to_snake_case(camel_str: str) -> str:

    s1 = _first_cap_re.sub(r'\1_\2', camel_str)
    return _all_cap_re.sub(r'\1_\2', s1).lower()


def snake_case_to_camel_case(snake_str: str) -> str:

    return re.sub(r"(?:^|_)(.)", lambda m: m.group(1).upper(), snake_str)


async def send_json_to_client(client: websockets.WebSocketServerProtocol, data: str) -> None:

    try:
        await client.send(data)
    except websockets.exceptions.ConnectionClosed:
        pass


async def server(client: Any,
                 path: str,
                 logger: Any,
                 register: Callable[[Any], Awaitable[None]],
                 unregister: Callable[[Any], Awaitable[None]],
                 rpc_dict: RPC_DICT_TYPE,
                 event_dict: Optional[EVENT_DICT_TYPE] = None,
                 verbose: bool = False) -> None:

    if event_dict is None:
        event_dict = {}

    req_last_ts: Dict[str, deque] = {}
    ignored_reqs: Set[str] = set()

    try:

        await register(client)

        async for message in client:

            try:
                data = json.loads(message)
            except json.decoder.JSONDecodeError as e:
                logger.error(f"Invalid data: '{message}'.")
                logger.debug(e)
                continue

            if not isinstance(data, dict):
                logger.error(f"Invalid data: '{data}'.")
                continue

            if "request" in data:  # ...then it is RPC

                try:
                    req_cls, resp_cls = RPC_MAPPING[data['request']]
                except KeyError:
                    logger.error(f"Unknown RPC request: {data}.")
                    continue

                if req_cls not in rpc_dict:
                    logger.debug(f"Ignoring RPC request: {data}.")
                    continue

                try:
                    req = req_cls.from_dict(data)
                except ValidationError as e:
                    logger.error(f"Invalid RPC: {data}, error: {e}")
                    continue

                try:
                    resp = await rpc_dict[req_cls](req, client)
                except Arcor2Exception as e:
                    logger.debug(e, exc_info=True)
                    resp = resp_cls()
                    resp.result = False
                    resp.messages = [e.message]

                if resp is None:  # default response
                    resp = resp_cls()
                else:
                    assert isinstance(resp, resp_cls)

                resp.id = req.id

                await asyncio.wait([client.send(resp.to_json())])

                if logger.level == LogLevel.DEBUG:

                    # Silencing of repetitive log messages
                    # ...maybe this could be done better and in a more general way using logging.Filter?

                    now = time.monotonic()
                    if req.request not in req_last_ts:
                        req_last_ts[req.request] = deque()

                    while req_last_ts[req.request]:
                        if req_last_ts[req.request][0] < now - 5.0:
                            req_last_ts[req.request].popleft()
                        else:
                            break

                    req_last_ts[req.request].append(now)
                    req_per_sec = len(req_last_ts[req.request]) / 5.0

                    if req_per_sec > 2:
                        if req.request not in ignored_reqs:
                            ignored_reqs.add(req.request)
                            logger.debug(f"Request of type {req.request} will be silenced.")
                    elif req_per_sec < 1:
                        if req.request in ignored_reqs:
                            ignored_reqs.remove(req.request)

                    if req.request not in ignored_reqs:
                        asyncio.ensure_future(logger.debug(f"RPC request: {req}, result: {resp}"))

            elif "event" in data:  # ...event from UI

                try:
                    event_cls = EVENT_MAPPING[data["event"]]
                except KeyError as e:
                    logger.error(f"Unknown event type: {e}.")
                    continue

                if event_cls not in event_dict:
                    logger.debug(f"Ignoring event: {data}.")
                    continue

                try:
                    event = event_cls.from_dict(data)
                except ValidationError as e:
                    logger.error(f"Invalid event: {data}, error: {e}")
                    continue

                await event_dict[event_cls](event, client)

            else:
                logger.error(f"unsupported format of message: {data}")
    except websockets.exceptions.ConnectionClosed:
        pass
    finally:
        await unregister(client)


def format_stacktrace() -> str:
    parts = ["Traceback (most recent call last):\n"]
    parts.extend(traceback.format_stack(limit=25)[:-2])
    parts.extend(traceback.format_exception(*sys.exc_info())[1:])
    return "".join(parts)


def print_exception(e: Exception) -> None:

    if isinstance(e, Arcor2Exception):
        pee = ProjectExceptionEvent(data=ProjectExceptionEventData(e.message, e.__class__.__name__, True))
    else:
        pee = ProjectExceptionEvent(data=ProjectExceptionEventData(str(e), e.__class__.__name__))

    print(pee.to_json())
    sys.stdout.flush()

    with open("traceback-{}.txt".format(time.strftime("%Y%m%d-%H%M%S")), "w") as tb_file:
        tb_file.write(format_stacktrace())


S = TypeVar('S')


async def run_in_executor(func: Callable[..., S], *args) -> S:
    return await asyncio.get_event_loop().run_in_executor(None, func, *args)


T = TypeVar('T')


def save_and_import_type_def(source: str, type_name: str, output_type: Type[T], path: str, module_name: str) -> Type[T]:
    """
    Save source to a file in object_type_path directory.

    :param source:
    :param type_name:
    :param output_type:
    :param object_type_path:
    :return:
    """

    if path not in sys.path:
        sys.path.append(path)

    type_file = camel_case_to_snake_case(type_name)
    full_path = os.path.join(path, module_name, type_file)

    with open(f"{full_path}.py", "w") as file:
        file.write(source)

    try:
        module = importlib.import_module(f"{module_name}.{type_file}")
    except ModuleNotFoundError:
        raise ImportClsException(f"Module '{module_name}' not found.")

    try:
        cls = getattr(module, type_name)
    except AttributeError:
        raise ImportClsException(f"Class {type_name} not found in module '{module_name}'.")

    if not issubclass(cls, output_type):
        raise ImportClsException("Not a required type.")

    return cls


def get_package_meta_path(package_id: str) -> str:

    return os.path.join(PROJECT_PATH, package_id, "package.json")


def read_package_meta(package_id: str) -> PackageMeta:

    try:
        with open(get_package_meta_path(package_id)) as pkg_file:
            return PackageMeta.from_json(pkg_file.read())
    except (IOError, ValidationError):
        return PackageMeta("N/A", datetime.fromtimestamp(0, tz=timezone.utc))


def write_package_meta(package_id: str, meta: PackageMeta) -> None:

    with open(get_package_meta_path(package_id), "w") as pkg_file:
        pkg_file.write(meta.to_json())


def check_compatibility(my_version: str, their_version: str) -> None:

    try:
        mv = semver.VersionInfo.parse(my_version)
        tv = semver.VersionInfo.parse(their_version)
    except ValueError as e:
        raise Arcor2Exception from e

    if mv.major != tv.major:
        raise Arcor2Exception("Different major varsion.")

    if mv.major == 0:
        if mv.minor != tv.minor:
            raise Arcor2Exception(f"Our version {my_version} is not compatible with {their_version}.")
    else:
        if mv.minor > tv.minor:
            raise Arcor2Exception(f"Our version {my_version} is outdated for {their_version}.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import asyncio
import shutil
from typing import Optional, Type

import horast

from arcor2 import helpers as hlp
from arcor2.action import patch_object_actions
from arcor2.clients import aio_persistent_storage as storage
from arcor2.data.object_type import ObjectModel, ObjectTypeMeta
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types import utils as otu
from arcor2.object_types.abstract import Generic, Robot
from arcor2.parameter_plugins import TYPE_TO_PLUGIN
from arcor2.server import globals as glob, settings
from arcor2.server.robot import get_robot_meta
from arcor2.source.object_types import prepare_object_types_dir


def get_obj_type_name(object_id: str) -> str:

    try:
        return glob.SCENE_OBJECT_INSTANCES[object_id].__class__.__name__
    except KeyError:
        raise Arcor2Exception("Unknown object id.")


def valid_object_types() -> otu.ObjectTypeDict:
    """
    To get only valid (not disabled) types.
    :return:
    """

    return {obj_type: obj for obj_type, obj in glob.OBJECT_TYPES.items() if not obj.meta.disabled}


def handle_robot_urdf(robot: Type[Robot]) -> None:

    if not robot.urdf_package_path:
        return

    shutil.copy(robot.urdf_package_path, settings.URDF_PATH)


async def get_object_data(object_types: otu.ObjectTypeDict, obj_id: str) -> None:

    if obj_id in object_types:
        return

    obj = await storage.get_object_type(obj_id)
    base = otu.base_from_source(obj.source, obj_id)
    if base and base not in object_types.keys():
        try:
            await get_object_data(object_types, base)
        except Arcor2Exception:
            object_types[obj_id] = \
                otu.ObjectTypeData(
                    ObjectTypeMeta(obj_id, "Object type disabled.", disabled=True, problem="Can't get base."))
            return

    try:
        type_def = await hlp.run_in_executor(
            hlp.save_and_import_type_def, obj.source, obj.id, Generic, settings.OBJECT_TYPE_PATH,
            settings.OBJECT_TYPE_MODULE)
        assert issubclass(type_def, Generic)
        meta = otu.meta_from_def(type_def)
    except Arcor2Exception as e:
        glob.logger.warning(f"Disabling object type {obj.id}.")
        glob.logger.debug(e, exc_info=True)
        object_types[obj_id] = \
            otu.ObjectTypeData(ObjectTypeMeta(obj_id, "Object type disabled.", disabled=True, problem=e.message))
        return

    patch_object_actions(type_def)

    if obj.model:
        model = await storage.get_model(obj.model.id, obj.model.type)
        kwargs = {model.type().value.lower(): model}
        meta.object_model = ObjectModel(model.type(), **kwargs)  # type: ignore

    ast = horast.parse(obj.source)
    otd = otu.ObjectTypeData(meta, type_def, otu.object_actions(TYPE_TO_PLUGIN, type_def, ast), ast)

    if issubclass(type_def, Robot):
        await get_robot_meta(otd)
        asyncio.ensure_future(hlp.run_in_executor(handle_robot_urdf, type_def))

    object_types[obj_id] = otd


async def get_object_types() -> None:

    await hlp.run_in_executor(prepare_object_types_dir, settings.OBJECT_TYPE_PATH, settings.OBJECT_TYPE_MODULE)

    object_types: otu.ObjectTypeDict = otu.built_in_types_data(TYPE_TO_PLUGIN)

    obj_ids = await storage.get_object_type_ids()

    for id_desc in obj_ids.items:
        await get_object_data(object_types, id_desc.id)

    for obj_type in object_types.values():

        # if description is missing, try to get it from ancestor(s)
        if not obj_type.meta.description:

            try:
                obj_type.meta.description = otu.obj_description_from_base(object_types, obj_type.meta)
            except otu.DataError as e:
                glob.logger.error(f"Failed to get info from base for {obj_type}, error: '{e}'.")

        if not obj_type.meta.disabled and not obj_type.meta.built_in:
            otu.add_ancestor_actions(obj_type.meta.type, object_types)

    glob.OBJECT_TYPES = object_types


async def get_robot_instance(robot_id: str, end_effector_id: Optional[str] = None) -> Robot:

    if robot_id not in glob.SCENE_OBJECT_INSTANCES:
        raise Arcor2Exception("Robot not found.")

    robot_inst = glob.SCENE_OBJECT_INSTANCES[robot_id]
    if not isinstance(robot_inst, Robot):
        raise Arcor2Exception("Not a robot.")
    if end_effector_id and end_effector_id not in await hlp.run_in_executor(robot_inst.get_end_effectors_ids):
        raise Arcor2Exception("Unknown end effector ID.")
    return robot_inst
